
"""Test the radiosity module with directional Patches."""
import os

import numpy as np
import numpy.testing as npt
import pyfar as pf
import pytest
import sparrowpy as sp
import sofar as sf


def check_energy_conservation(
        receivers, receiver_weights, data, absorption_data=0):
    integration_weights = receiver_weights.copy()
    integration_weights *= 2 * np.pi / np.sum(integration_weights)
    integration_weights *= np.cos(receivers.colatitude)
    integration_weights = integration_weights[..., np.newaxis]
    energy = np.sum(data.freq * integration_weights, axis=1)
    npt.assert_almost_equal(
        energy, 1 - absorption_data, decimal=1)


def check_reciprocity(data):
    data_array = data.freq
    for i_source in range(data_array.shape[0]):
        for i_receiver in range(i_source + 1, data_array.shape[1]):
            npt.assert_almost_equal(
                data_array[i_source, i_receiver],
                data_array[i_receiver, i_source])


def test_create_from_scattering_with_valid_data(tmp_path):
    # Prepare test data
    scattering_data = [0, 0.2, 1]
    frequency_data = [100, 200, 400]
    file_path = os.path.join(tmp_path, "test_brdf.sofa")
    coords = pf.samplings.sph_gaussian(sh_order=1)
    coords = coords[coords.z > 0]

    # Call the function
    sp.brdf.create_from_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data, frequency_data),
        file_path=file_path,
        )
    data, s, r = pf.io.read_sofa(file_path)

    # test coords
    npt.assert_almost_equal(s.cartesian, coords.cartesian)
    npt.assert_almost_equal(r.cartesian, coords.cartesian)
    # Assert the expected outcome
    assert data.freq.shape == (4, 4, 3)
    assert data.freq.shape == (s.csize, r.csize, 3)
    # test
    for i in range(4):
        for j in range(4):
            npt.assert_almost_equal(data.freq[i, j], data.freq[j, i])


def test_create_from_scattering_1(tmp_path):
    # Prepare test data
    scattering_data = np.ones((3, ))
    frequency_data = [100, 200, 400]
    file_path = os.path.join(tmp_path, "test_brdf.sofa")
    coords = pf.samplings.sph_gaussian(sh_order=11)
    coords = coords[coords.z > 0]

    # Call the function
    sp.brdf.create_from_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data, frequency_data),
        file_path=file_path,
        )
    data, s, r = pf.io.read_sofa(file_path)

    # Assert the expected outcome
    assert data.freq.shape == (s.csize, r.csize, 3)
    # test
    npt.assert_almost_equal(data.freq, 1 / np.pi)

    check_energy_conservation(
        coords, coords.weights, data)
    check_reciprocity(data)


def test_create_from_scattering_0(tmp_path):
    # Prepare test data
    scattering_data = np.zeros((3, ))
    frequency_data = [100, 200, 400]
    file_path = os.path.join(tmp_path, "test_brdf.sofa")
    coords = pf.samplings.sph_gaussian(sh_order=11)
    coords = coords[coords.z > 0]

    # Call the function
    sp.brdf.create_from_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data, frequency_data),
        file_path=file_path,
        )
    data, s, r = pf.io.read_sofa(file_path)

    # Assert the expected outcome
    assert data.freq.shape == (s.csize, r.csize, 3)
    # tests if energy is added to the specular direction
    coords_spec = coords.copy()
    coords_spec.azimuth += np.pi
    idx = coords.find_nearest(coords_spec)[0]
    npt.assert_array_less(0, data.freq[np.arange(coords.csize), idx])

    check_energy_conservation(
        coords, coords.weights, data)
    check_reciprocity(data)


def test_create_from_scattering_0_3(tmp_path):
    # Prepare test data
    scattering_data = 0.3
    frequency_data = [100, 200, 400]
    file_path = os.path.join(tmp_path, "test_brdf.sofa")
    coords = pf.samplings.sph_gaussian(sh_order=11)
    coords = coords[coords.z > 0]

    # Call the function
    sp.brdf.create_from_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data + np.zeros((3, )), frequency_data),
        file_path=file_path,
        )
    data, s, r = pf.io.read_sofa(file_path)

    # Assert the expected outcome
    assert data.freq.shape == (s.csize, r.csize, 3)

    check_energy_conservation(
        coords, coords.weights, data)
    check_reciprocity(data)


def test_create_from_scattering_0_3_with_absorption(tmp_path):
    # Prepare test data
    scattering_data = 0.3
    absorption_data = 0.3
    frequency_data = [100, 200, 400]
    file_path = os.path.join(tmp_path, "test_brdf.sofa")
    coords = pf.samplings.sph_gaussian(sh_order=1)
    coords = coords[coords.z > 0]

    # Call the function
    sp.brdf.create_from_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data + np.zeros((3, )), frequency_data),
        pf.FrequencyData(absorption_data + np.zeros((3, )), frequency_data),
        file_path=file_path,
        )
    data, s, r = pf.io.read_sofa(file_path)

    # Assert the expected outcome
    assert data.freq.shape == (s.csize, r.csize, 3)

    check_energy_conservation(
        coords, coords.weights, data, absorption_data)
    check_reciprocity(data)


def test_create_from_scattering_with_invalid_data(tmp_path):
    # Prepare test data
    scattering_data = pf.FrequencyData(1, 100)
    coords = pf.samplings.sph_gaussian(sh_order=1)

    # Call the function and expect it to raise an error
    with pytest.raises(
            TypeError,
            match="scattering_coefficient must be a pf.FrequencyData object"):
        sp.brdf.create_from_scattering(
            coords, coords, 'invalid', file_path=tmp_path)
    # Call the function and expect it to raise an error
    with pytest.raises(
            TypeError,
            match="source_directions must be a pf.Coordinates object"):
        sp.brdf.create_from_scattering(
            'coords', coords, scattering_data, file_path=tmp_path)
    with pytest.raises(
            TypeError,
            match="receiver_directions must be a pf.Coordinates object"):
        sp.brdf.create_from_scattering(
            coords, 'coords', scattering_data, file_path=tmp_path)


@pytest.mark.parametrize('scattering_data', [0, 0.3, 1])
@pytest.mark.parametrize('absorption_data', [0, 0.3, 1])
@pytest.mark.parametrize('sh_order', [5, 11])
def test_create_from_scattering_energy_conservation(
        tmp_path, scattering_data, absorption_data, sh_order):
    # Prepare test data
    frequency_data = [100, 200, 400]
    file_path = os.path.join(tmp_path, "test_brdf.sofa")
    coords = pf.samplings.sph_gaussian(sh_order=sh_order)
    coords = coords[coords.z > 0]

    # Call the function
    sp.brdf.create_from_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data + np.zeros((3, )), frequency_data),
        pf.FrequencyData(absorption_data + np.zeros((3, )), frequency_data),
        file_path=file_path,
        )
    sofa = sf.read_sofa(file_path)
    data, s, r = pf.io.convert_sofa(sofa)

    # test energy conservation
    check_energy_conservation(r, sofa.ReceiverWeights, data, absorption_data)


@pytest.mark.parametrize('scattering_data', [0, 0.3, 1])
@pytest.mark.parametrize('absorption_data', [0, 0.3, 1])
@pytest.mark.parametrize('sh_order', [5, 11])
def test_create_from_scattering_reciprocal_principle(
        tmp_path, scattering_data, absorption_data, sh_order):
    # Prepare test data
    frequency_data = [100, 200, 400]
    file_path = os.path.join(tmp_path, "test_brdf.sofa")
    coords = pf.samplings.sph_gaussian(sh_order=sh_order)
    coords = coords[coords.z > 0]

    # Call the function
    sp.brdf.create_from_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data + np.zeros((3, )), frequency_data),
        pf.FrequencyData(absorption_data + np.zeros((3, )), frequency_data),
        file_path=file_path,
        )
    sofa = sf.read_sofa(file_path)
    data, s, r = pf.io.convert_sofa(sofa)

    # test energy conservation
    check_reciprocity(data)


@pytest.mark.parametrize('absorption_data', [0, 0.5])
@pytest.mark.parametrize('sh_order', [5, 11])
def test_directional_energy_conservation_specular(
        tmp_path, absorption_data, sh_order):
    # Prepare test data
    frequency_data = [100, 200, 400]
    file_path = os.path.join(tmp_path, "test_brdf.sofa")
    coords = pf.samplings.sph_gaussian(sh_order=sh_order)
    coords = coords[coords.z > 0]
    scattering_data = np.zeros((coords.csize, coords.csize, 3))
    scattering_data[np.arange(coords.csize), np.arange(coords.csize)] = 1
    npt.assert_array_equal(np.sum(scattering_data, axis=1), 1)

    # Call the function
    sp.brdf.create_from_directional_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data, frequency_data),
        pf.FrequencyData(absorption_data + np.zeros((3, )), frequency_data),
        file_path=file_path,
        )
    sofa = sf.read_sofa(file_path)
    data, _s, r = pf.io.convert_sofa(sofa)

    # test energy conservation and reciprocity
    check_energy_conservation(
        r, sofa.ReceiverWeights, data, absorption_data)
    check_reciprocity(data)


@pytest.mark.parametrize('absorption_data', [0, 0.3, 1])
@pytest.mark.parametrize('sh_order', [7, 11])
def test_directional_energy_conservation_diffuse(
        absorption_data, sh_order):
    # Prepare test data
    frequency_data = [100, 200, 400]
    coords = pf.samplings.sph_gaussian(sh_order=sh_order)
    coords = coords[coords.z > 0]
    coords.weights *= 2 * np.pi / np.sum(coords.weights)
    cos_receiver = np.cos(coords.colatitude)[np.newaxis, :, np.newaxis]
    weights_receiver = coords.weights[np.newaxis, :, np.newaxis]
    scattering_data = cos_receiver * weights_receiver / np.pi *  np.ones(
        (coords.csize, coords.csize, 3))

    npt.assert_almost_equal(np.sum(scattering_data, axis=1), 1, decimal=2)

    # Call the function
    brdf = sp.brdf.create_from_directional_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data, frequency_data),
        pf.FrequencyData(absorption_data + np.zeros((3, )), frequency_data),
        )

    # test energy conservation and reciprocity
    check_energy_conservation(
        coords, coords.weights, brdf, absorption_data)
    check_reciprocity(brdf)

    npt.assert_almost_equal(brdf.freq, (1-absorption_data) / np.pi, decimal=2)


@pytest.mark.parametrize('absorption_data', [0, 0.3, 1])
@pytest.mark.parametrize('sh_order', [5, 11])
def test_directional_reciprocal_principle_specular(
        tmp_path, absorption_data, sh_order):
    # Prepare test data
    frequency_data = [100, 200, 400]
    file_path = os.path.join(tmp_path, "test_brdf.sofa")
    coords = pf.samplings.sph_gaussian(sh_order=sh_order)
    coords = coords[coords.z > 0]
    scattering_data = np.zeros((coords.csize, coords.csize, 3))
    scattering_data[np.arange(coords.csize), np.arange(coords.csize)] = 1
    npt.assert_array_equal(np.sum(scattering_data, axis=1), 1)

    # Call the function
    sp.brdf.create_from_directional_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data, frequency_data),
        pf.FrequencyData(absorption_data + np.zeros((3, )), frequency_data),
        file_path=file_path,
        )
    sofa = sf.read_sofa(file_path)
    data, _, r = pf.io.convert_sofa(sofa)

    # test energy conservation and reciprocity
    check_energy_conservation(
        r, sofa.ReceiverWeights, data, absorption_data)
    check_reciprocity(data)


@pytest.mark.parametrize('absorption_data', [0, 0.5])
@pytest.mark.parametrize('sh_order', [5, 11])
def test_directional_energy_conservation_two_directions(
        tmp_path, absorption_data, sh_order):
    # Prepare test data
    frequency_data = [100, 200, 400]
    file_path = os.path.join(tmp_path, "test_brdf.sofa")
    coords = pf.samplings.sph_gaussian(sh_order=sh_order)
    coords = coords[coords.z > 0]
    scattering_data = np.zeros((coords.csize, coords.csize, 3))
    scattering_data[np.arange(coords.csize), np.arange(coords.csize)] = 0.5
    refl = coords.copy()
    refl.azimuth += np.pi
    idx = coords.find_nearest(refl)[0]
    scattering_data[np.arange(coords.csize), idx] = 0.5
    npt.assert_array_equal(np.sum(scattering_data, axis=1), 1)

    # Call the function
    sp.brdf.create_from_directional_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data, frequency_data),
        pf.FrequencyData(absorption_data + np.zeros((3, )), frequency_data),
        file_path=file_path,
        )
    sofa = sf.read_sofa(file_path)
    data, _s, r = pf.io.convert_sofa(sofa)

    # test energy conservation and reciprocity
    check_energy_conservation(
        r, sofa.ReceiverWeights, data, absorption_data)
    check_reciprocity(data)


@pytest.mark.parametrize('absorption_data', [0, 0.5])
@pytest.mark.parametrize('sh_order', [5, 11])
def test_directional_energy_conservation_two_directions_rand(
        tmp_path, absorption_data, sh_order):
    # Prepare test data
    frequency_data = [100, 200, 400]
    coords = pf.samplings.sph_gaussian(sh_order=sh_order)
    coords = coords[coords.z > 0]
    scattering_data = np.zeros((coords.csize, coords.csize, 3))
    scattering_data[
        np.arange(coords.csize), np.arange(coords.csize)] = 0.5
    specular_coords = coords.copy()
    specular_coords.azimuth += np.pi
    idx = coords.find_nearest(specular_coords)[0]
    scattering_data[
        np.arange(coords.csize), idx] += 0.5
    npt.assert_array_equal(np.sum(scattering_data, axis=1), 1)

    # Call the function
    file_path = os.path.join(tmp_path, "test_brdf.sofa")
    sp.brdf.create_from_directional_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data, frequency_data),
        pf.FrequencyData(absorption_data + np.zeros((3, )), frequency_data),
        file_path=file_path,
        )
    sofa = sf.read_sofa(file_path)
    data, _s, r = pf.io.convert_sofa(sofa)

    # test energy conservation and reciprocity
    check_energy_conservation(
        r, sofa.ReceiverWeights, data, absorption_data)
    check_reciprocity(data)


@pytest.mark.parametrize('absorption_data', [0, 0.5])
@pytest.mark.parametrize('sh_order', [5, 11])
def test_directional_random_directional_scattering(
        tmp_path, absorption_data, sh_order):
    # Prepare test data
    frequency_data = [100, 200, 400]
    file_path = os.path.join(tmp_path, "test_brdf.sofa")
    coords = pf.samplings.sph_gaussian(sh_order=sh_order)
    coords = coords[coords.z > 0]
    rng = np.random.default_rng(1337)
    scattering_data = rng.random((coords.csize, coords.csize, 3))
    scattering_data /= np.sum(scattering_data, axis=1, keepdims=True)
    npt.assert_almost_equal(np.sum(scattering_data, axis=1), 1)

    # Call the function
    sp.brdf.create_from_directional_scattering(
        coords, coords,
        pf.FrequencyData(scattering_data, frequency_data),
        pf.FrequencyData(absorption_data + np.zeros((3, )), frequency_data),
        file_path=file_path,
        )
    sofa = sf.read_sofa(file_path)
    data, _, r = pf.io.convert_sofa(sofa)

    # test energy conservation and reciprocity
    check_energy_conservation(
        r, sofa.ReceiverWeights, data, absorption_data)



"""Stub utilities for testing."""
import sparrowpy as sp


def shoebox_room_stub(length_x, length_y, length_z):
    """Create a shoebox room with the given dimensions.

    Parameters
    ----------
    length_x : float
        Length of the room in x-direction in meters.
    length_y : float
        Length of the room in y-direction in meters
    length_z : float
        Length of the room in z-direction in meters

    Returns
    -------
    room : list[geo.Polygon]
        List of the walls of the room.

    """
    return [
        sp.geometry.Polygon(
            [[0, 0, 0], [length_x, 0, 0],
            [length_x, 0, length_z], [0, 0, length_z]],
            [1, 0, 0], [0, 1, 0]),
        sp.geometry.Polygon(
            [[0, length_y, 0], [length_x, length_y, 0],
            [length_x, length_y, length_z], [0, length_y, length_z]],
            [1, 0, 0], [0, -1, 0]),
        sp.geometry.Polygon(
            [[0, 0, 0], [length_x, 0, 0],
            [length_x, length_y, 0], [0, length_y, 0]],
            [1, 0, 0], [0, 0, 1]),
        sp.geometry.Polygon(
            [[0, 0, length_z], [length_x, 0, length_z],
            [length_x, length_y, length_z], [0, length_y, length_z]],
            [1, 0, 0], [0, 0, -1]),
        sp.geometry.Polygon(
            [[0, 0, 0], [0, 0, length_z],
            [0, length_y, length_z], [0, length_y, 0]],
            [0, 0, 1], [1, 0, 0]),
        sp.geometry.Polygon(
            [[length_x, 0, 0], [length_x, 0, length_z],
            [length_x, length_y, length_z], [length_x, length_y, 0]],
            [0, 0, 1], [-1, 0, 0]),
        ]



"""
This files validates the Radiosity method for a diffuse infinite plane.

The analytical results are taken from Svensson et. al. [1].

[1] U. P. Svensson and L. Savioja, "The Lambert diffuse reflection model
revisited," The Journal of the Acoustical Society of America, vol. 156,
no. 6, pp. 3788-3796, Dec. 2024, doi: 10.1121/10.0034561.

"""

import numpy as np
import sparrowpy as sp
import pyfar as pf
import pytest
import numpy.testing as npt


def run_energy_diff_specular_ratio(
        width, length, patch_size, source, receiver):
    """
    Calculate the ratio of diffuse to specular energy for an plane.
    The plane is located in the x-y plane. Its center is at (0, 0, 0).

    Parameters
    ----------
    width : float
        Width of the plane.
    length : float
        length of the plane.
    patch_size : float
        Size of the patches.
    source : pf.Coordinates
        Position of the source.
    receiver : pf.Coordinates
        Position of the receiver in cartesian.

    Returns
    -------
    ratio : float
        Ratio of diffuse to specular energy.
    """
    source_is = source.copy()
    source_is.z *= -1
    reflection_len =  (receiver - source_is).radius[0]
    speed_of_sound = 343
    sampling_rate = 1
    etc_duration = reflection_len/speed_of_sound
    etc_duration=1

    plane = sp.geometry.Polygon(
            [[-width/2, -length/2, 0],
             [width/2, -length/2, 0],
             [width/2, length/2, 0],
             [-width/2, length/2, 0]],
            [1, 0, 0], [0, 0, 1])

    #simulation parameters
    radi = sp.DirectionalRadiosityFast.from_polygon(
        [plane], patch_size)

    brdf_sources = pf.Coordinates(0, 0, 1, weights=1)
    brdf_receivers = pf.Coordinates(0, 0, 1, weights=1)
    brdf = sp.brdf.create_from_scattering(
        brdf_sources,
        brdf_receivers,
        pf.FrequencyData(1, [100]),
        pf.FrequencyData(0, [100]),
    )

    radi.set_wall_brdf(
        np.arange(1), brdf, brdf_sources, brdf_receivers)

    # set air absorption
    radi.set_air_attenuation(
        pf.FrequencyData(
            np.zeros_like(brdf.frequencies),
            brdf.frequencies))

    # initialize source energy at each patch
    radi.init_source_energy(source)

    # gather energy at receiver
    radi.calculate_energy_exchange(
        speed_of_sound=speed_of_sound,
        etc_time_resolution=1/sampling_rate,
        etc_duration=etc_duration,
        max_reflection_order=0)

    I_diffuse = radi.collect_energy_receiver_mono(receiver)

    I_specular = 1/(4*np.pi*reflection_len**2)
    return np.sum(I_diffuse.time)/I_specular


@pytest.mark.parametrize("patch_size", [1])
def test_colocated_source_receiver(patch_size):
    """
    Compares test case 1 from [1], where source and receiver are located
    at the same position. For this case the Energy ratio diffuse to
    specular should be 2.
    """
    width = 20
    depth = 20
    source = pf.Coordinates(0, 0, 3, weights=1)
    receiver = pf.Coordinates(0, 0, 3, weights=1)
    ratio = run_energy_diff_specular_ratio(
        width, depth, patch_size, source, receiver)

    # the energy should be 2 ideally, but the simulation cannot cover
    # an infinite plane, therefore the energy is slightly less than 2
    npt.assert_array_less(ratio, 2)
    npt.assert_allclose(ratio, 1.97, rtol=0.01)


@pytest.mark.parametrize("patch_size", [1])
def test_source_receiver_along_same_normal(patch_size):
    """
    Compares test case 2 from [1], where Source and receiver are along
    the same normal. For this case the Energy ratio diffuse to
    specular should be 2.
    """
    width = 20
    depth = 20
    source = pf.Coordinates(0, 0, 5, weights=1)
    receiver = pf.Coordinates(0, 0, 3, weights=1)
    ratio = run_energy_diff_specular_ratio(
        width, depth, patch_size, source, receiver)

    # the energy should be 2 ideally, but the simulation cannot cover
    # an infinite plane, therefore the energy is slightly less than 2
    npt.assert_array_less(ratio, 2)
    npt.assert_allclose(ratio, 1.97, rtol=0.01)


@pytest.mark.parametrize("theta_deg", [30, 45, 60])
@pytest.mark.parametrize("patch_size", [1])
def test_source_receiver_same_hight(patch_size, theta_deg):
    """
    Compares test case 3 from [1], where receiver is at the same height as
    the source, but moved sideways/laterally. For this case the Energy
    ratio diffuse to specular should be 2cos(theta). With theta being the
    incident angle of the specular reflection.
    """
    width = 20
    depth = 20
    theta_rad = np.deg2rad(theta_deg)
    source = pf.Coordinates.from_spherical_colatitude(
        0, theta_rad, 2, weights=1)
    receiver = pf.Coordinates.from_spherical_colatitude(
        np.pi, theta_rad, 2, weights=1)
    ratio = run_energy_diff_specular_ratio(
        width, depth, patch_size, source, receiver)

    npt.assert_allclose(ratio, 2*np.cos(theta_rad), atol=0.03, rtol=0.03)




"""Test the radiosity_sound_object module."""
import os

import numpy as np
import numpy.testing as npt
import pyfar as pf
import pytest
import sparrowpy.sound_object as so


def test_directivity_ms():
    """Test DirectivityMS class."""
    path = os.path.join(
        os.path.dirname(__file__), 'test_data', 'ITA_Dodecahedron.sofa')
    directivity = so.DirectivityMS(path)
    assert directivity.data.cshape[0] == directivity.receivers.cshape[0]
    assert isinstance(directivity.data, pf.FrequencyData)
    assert isinstance(directivity.receivers, pf.Coordinates)


def test_get_directivity():
    """Test get_directivity function."""
    path = os.path.join(
        os.path.dirname(__file__), 'test_data', 'ITA_Dodecahedron.sofa')
    directivity = so.DirectivityMS(path)
    target_pos_G = [-1, 0, -2]
    pos_G = [0, 0, 2]
    up_G = [0, 1, 0]
    view_G = [-0.707106781186548, 0, 0.707106781186548]
    fac = directivity.get_directivity(pos_G, view_G, up_G, target_pos_G, 2)
    npt.assert_almost_equal(fac, 0.0001908)


@pytest.mark.parametrize('function', [
    (so.SoundObject),
    (so.SoundSource),
    (so.Receiver) ])
def test_sound_object(function):
    """Test SoundObject class."""
    sound_object = function([0, 0, 0], [2, 0, 0], [0, 0, 2])
    npt.assert_equal(sound_object.position, np.array([0, 0, 0]))
    npt.assert_equal(sound_object.view, np.array([1, 0, 0]))
    npt.assert_equal(sound_object.up, np.array([0, 0, 1]))


@pytest.mark.parametrize('function', [
    (so.SoundObject),
    (so.SoundSource),
    (so.Receiver) ])
@pytest.mark.parametrize(('position', 'view', 'up'), [
    ('a', [1, 0, 0], [0, 0, 1]),
    ([0, 0, 0], [1, 0, 0], 1),
    ([0, 0, 0], ['a'], [0, 0, 1]),
])
def test_sound_object_error(function, position, view, up):
    """Test SoundObject class."""
    with pytest.raises((ValueError, AssertionError)):
        function(position, view, up)


def test_sound_source_defaults():
    """Test SoundSource class."""
    sound_source = so.SoundSource([0, 0, 0], [2, 0, 0], [0, 0, 2])
    npt.assert_equal(sound_source.sound_power, 1)
    npt.assert_equal(sound_source.directivity, None)


@pytest.mark.parametrize(('directivity', 'sound_power'), [
    ('a', 1),
    (None, 'a'),
])
def test_sound_source_errors(directivity, sound_power):
    """Test SoundSource class."""
    with pytest.raises((ValueError, AssertionError)):
        so.SoundSource(
            [0, 0, 0], [2, 0, 0], [0, 0, 2], directivity, sound_power)


def test__get_metrics():
    """Test _get_metrics function."""
    target_pos_G = [-1, 0, -2]
    pos_G = [0, 0, 2]
    up_G = [0, 1, 0]
    view_G = [-0.707106781186548, 0, 0.707106781186548]
    (azimuth_deg, elevation_deg) = so._get_metrics(
        pos_G, view_G, up_G, target_pos_G)
    assert azimuth_deg == -120.96375653207352
    assert elevation_deg == 0



"""Test radiosity module."""
import numpy as np
import numpy.testing as npt
import pytest
import pyfar as pf
import os
import sparrowpy as sp
import sofar as sf

def test_init(sample_walls):
    radiosity = sp.DirectionalRadiosityFast.from_polygon(sample_walls, 0.5)
    npt.assert_almost_equal(radiosity.patches_points.shape, (24, 4, 3))
    npt.assert_almost_equal(radiosity.patches_area.shape, (24))
    npt.assert_almost_equal(radiosity.patches_center.shape, (24, 3))
    npt.assert_almost_equal(radiosity.patches_size.shape, (24, 3))
    npt.assert_almost_equal(radiosity.patches_normal.shape, (24, 3))


def test_check_visibility(sample_walls):
    radiosity = sp.DirectionalRadiosityFast.from_polygon(sample_walls, .5)
    radiosity.bake_geometry()
    npt.assert_almost_equal(radiosity._visibility_matrix.shape, (24,24))
    for i in range(6):
        k = i*4
        npt.assert_array_equal(radiosity._visibility_matrix[k+4:, k:k+4],
                               False)
        npt.assert_array_equal(radiosity._visibility_matrix[k:k+4, k+4:],
                               True)
    assert np.sum(radiosity._visibility_matrix) == 24**2-21*4**2


def test_check_visibility_wrapper(sample_walls):
    radiosity = sp.DirectionalRadiosityFast.from_polygon(sample_walls, 0.5)
    radiosity.bake_geometry()
    visibility_matrix = sp.geometry._check_patch2patch_visibility(
        radiosity.patches_center, radiosity.patches_normal,
        radiosity.patches_points)
    npt.assert_almost_equal(radiosity._visibility_matrix, visibility_matrix)


def test_compute_form_factors(sample_walls):
    radiosity = sp.DirectionalRadiosityFast.from_polygon(sample_walls, .5)
    radiosity.bake_geometry()
    npt.assert_almost_equal(radiosity.form_factors.shape, (24, 24))

def test_patch_2_out_dir_mapping():
    """Test patch centroid to brdf receiver direction map."""

    # input: two orthogonal walls with flipped up vector
    p0 = [[0,0,0],[0,1,0],[0,1,1],[0,0,1]]
    n0=[1,0,0]
    u0=[0,0,1]
    p1 = [[0,0,0],[1,0,0],[1,0,1],[0,0,1]]
    n1=[0,1,0]
    u1=[0,0,-1]

    walls = [sp.geometry.Polygon(points=p0,normal=n0,up_vector=u0),
             sp.geometry.Polygon(points=p1,normal=n1,up_vector=u1)]

    radiosity = sp.DirectionalRadiosityFast.from_polygon(walls, 1)

    # set brdf sampling with 45º resolution:
    # ensure predictable positions and some outgoing directions in horiz. plane
    samples = pf.samplings.sph_equal_angle(delta_angles=45)
    samples.weights=np.ones(samples.cshape[0])

    brdf_sources = samples[np.where((samples.elevation*180/np.pi >= 0))].copy()
    brdf_receivers=samples[np.where((samples.elevation*180/np.pi >= 0))].copy()
    frequencies = np.array([1000])

    brdf = sp.brdf.create_from_scattering(
        brdf_sources,
        brdf_receivers,
        pf.FrequencyData(.5*np.ones_like(frequencies), frequencies),
        pf.FrequencyData(.1*np.ones_like(frequencies), frequencies))

    radiosity.set_wall_brdf(
    np.arange(len(walls)), brdf, brdf_sources, brdf_receivers)

    radiosity.bake_geometry()

    # ensure that mapping has correct dimensions
    assert radiosity._patch_2_brdf_outgoing_index.ndim==2
    assert radiosity._patch_2_brdf_outgoing_index.shape[0]==radiosity.n_patches
    assert radiosity._patch_2_brdf_outgoing_index.shape[1]==radiosity.n_patches

    # index of own centroid stores invalid entry
    assert (radiosity._patch_2_brdf_outgoing_index[0,0] ==
            radiosity._brdf_outgoing_directions[0].cshape[0])
    assert (radiosity._patch_2_brdf_outgoing_index[1,1] ==
            radiosity._brdf_outgoing_directions[0].cshape[0])

    # index of centroid 1 relative to patch 0
    i0 = radiosity._patch_2_brdf_outgoing_index[0,1]
    # index of centroid 1 relative to patch 0
    i1 = radiosity._patch_2_brdf_outgoing_index[1,0]

    # indices are the same because the up vectors are flipped
    # there is a symmetry over the x=y plane
    assert (i0 == i1)
    v0=radiosity._brdf_outgoing_directions[0].cartesian[int(i0)]
    v1=radiosity._brdf_outgoing_directions[1].cartesian[int(i1)]
    # corresponding directions in xy plane
    npt.assert_almost_equal(v0[2],0)
    npt.assert_almost_equal(v1[2],0)
    # corresponding directions are symmetrical
    npt.assert_almost_equal(v0,-v1)


@pytest.mark.parametrize('walls', [
    # perpendicular walls
    [0, 2], [0, 3], [0, 4], [0, 5],
    [1, 2], [1, 3], [1, 4], [1, 5],
    [2, 0], [2, 1], [2, 4], [2, 5],
    [3, 0], [3, 1], [3, 4], [3, 5],
    [4, 0], [4, 1], [4, 2], [4, 3],
    [5, 0], [5, 1], [5, 2], [5, 3],
    # parallel walls
    [0, 1], [2, 3], [4, 5],
    [1, 0], [3, 2], [5, 4],
    ])
@pytest.mark.parametrize('patch_size', [
    0.5,
    ])
def test_form_factors_directivity_for_diffuse(
        sample_walls, walls, patch_size, sofa_data_diffuse):
    """Test form factor calculation for perpendicular walls."""
    wall_source = sample_walls[walls[0]]
    wall_receiver = sample_walls[walls[1]]
    walls = [wall_source, wall_receiver]

    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        walls, patch_size)
    data, sources, receivers = sofa_data_diffuse
    radiosity.set_wall_brdf(
        np.arange(len(walls)), data, sources, receivers)
    radiosity.set_air_attenuation(
        pf.FrequencyData(np.zeros_like(data.frequencies), data.frequencies))
    radiosity.bake_geometry()

    form_factors_from_tilde = np.max(radiosity._form_factors_tilde, axis=2)
    for i in range(4):
        npt.assert_almost_equal(
            radiosity.form_factors[:4, 4:], form_factors_from_tilde[:4, 4:, i])
        npt.assert_almost_equal(
            radiosity.form_factors[:4, 4:].T,
            form_factors_from_tilde[4:, :4, i])

    for i in range(8):
        npt.assert_almost_equal(radiosity._form_factors_tilde[i, i, :, :], 0)


def test_set_wall_scattering(sample_walls, sofa_data_diffuse):
    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        sample_walls, 0.2)
    (data, sources, receivers) = sofa_data_diffuse
    radiosity.set_wall_brdf(np.arange(6), data, sources, receivers)
    # check shape of scattering matrix
    assert len(radiosity._brdf) == 1
    npt.assert_almost_equal(radiosity._brdf[0].shape, (4, 4, 4))
    npt.assert_array_equal(radiosity._brdf[0], 1)
    npt.assert_array_equal(radiosity._brdf_index, 0)
    # check source and receiver direction
    for i in range(6):
        assert (np.sum(
            radiosity._brdf_incoming_directions[i].cartesian * \
                radiosity.walls_normal[i,:],
            axis=-1)>0).all()
        assert (np.sum(
            radiosity._brdf_outgoing_directions[i].cartesian * \
                radiosity.walls_normal[i,:],
            axis=-1)>0).all()


def test_set_wall_scattering_different(sample_walls, sofa_data_diffuse):
    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        sample_walls, 0.2)
    (data, sources, receivers) = sofa_data_diffuse
    radiosity.set_wall_brdf([0, 1, 2], data, sources, receivers)
    radiosity.set_wall_brdf([3, 4, 5], data, sources, receivers)
    # check shape of scattering matrix
    assert len(radiosity._brdf) == 2
    for i in range(2):
        npt.assert_almost_equal(radiosity._brdf[i].shape, (4, 4, 4))
        npt.assert_array_equal(radiosity._brdf[i], 1)
    npt.assert_array_equal(radiosity._brdf_index[:3], 0)
    npt.assert_array_equal(radiosity._brdf_index[3:], 1)
    # check source and receiver direction
    for i in range(6):
        assert (np.sum(
            radiosity._brdf_incoming_directions[i].cartesian * \
                radiosity.walls_normal[i,:],
            axis=-1)>0).all()
        assert (np.sum(
            radiosity._brdf_outgoing_directions[i].cartesian * \
                radiosity.walls_normal[i,:],
            axis=-1)>0).all()


def test_set_air_attenuation(sample_walls):
    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        sample_walls, 0.2)
    radiosity.set_air_attenuation(pf.FrequencyData([0.1, 0.2], [500, 1000]))
    npt.assert_array_equal(radiosity._air_attenuation, [0.1, 0.2])


def test_total_number_of_patches():
    points = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]])
    result = sp.geometry._total_number_of_patches(points, 0.2)
    desired = 25
    assert result == desired


def test_init_source_with_directivity(sample_walls):
    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        sample_walls, 1)

    position = np.array([0, 0, 0])
    view = np.array([1, 0, 0])
    up = np.array([0, 0, 1])
    path = os.path.join(
        os.path.dirname(__file__), 'test_data',
        'Genelec8020_DAF_2016_1x1.v17.ms.sofa')
    directivity = sp.sound_object.DirectivityMS(path)
    sound_source = sp.sound_object.SoundSource(position, view, up, directivity)

    # try without directivity
    radiosity.init_source_energy(pf.Coordinates(*position))
    init_energy = radiosity._energy_init_source
    distance_patches_to_source = radiosity._distance_patches_to_source
    assert isinstance(radiosity._source, pf.Coordinates)

    # try with directivity
    radiosity.init_source_energy(sound_source)
    init_energy_dir = radiosity._energy_init_source
    distance_patches_to_source_dir = radiosity._distance_patches_to_source

    assert not np.allclose(init_energy, init_energy_dir)
    npt.assert_allclose(
        distance_patches_to_source, distance_patches_to_source_dir)
    assert isinstance(radiosity._source, sp.sound_object.SoundSource)


def test_init_source_with_directivity_frequency(sample_walls):
    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        sample_walls, 1)

    position = np.array([0, 0, 0])
    view = np.array([1, 0, 0])
    up = np.array([0, 0, 1])
    path = os.path.join(
        os.path.dirname(__file__), 'test_data',
        'Genelec8020_DAF_2016_1x1.v17.ms.sofa')

    # create directivity
    directivity = sp.sound_object.DirectivityMS(path)
    sound_source = sp.sound_object.SoundSource(position, view, up, directivity)

    # set air attenuation
    frequencies = sf.read_sofa(path, verbose=False).N
    radiosity.set_air_attenuation(
        pf.FrequencyData(np.zeros_like(frequencies), frequencies))

    # try without directivity
    radiosity.init_source_energy(pf.Coordinates(*position))
    init_energy = radiosity._energy_init_source
    distance_patches_to_source = radiosity._distance_patches_to_source

    # try with directivity
    radiosity.init_source_energy(sound_source)
    init_energy_dir = radiosity._energy_init_source
    distance_patches_to_source_dir = radiosity._distance_patches_to_source

    assert not np.allclose(init_energy, init_energy_dir)
    npt.assert_allclose(
        distance_patches_to_source, distance_patches_to_source_dir)

    # test if first frequency bin is same as above
    npt.assert_allclose(
        np.array([
            [0.        ], [2.13215939], [0.        ],
            [2.16329116], [0.        ], [2.12899306]]),
            init_energy_dir[..., 0])

    # test if other frequency bins are not same as first one
    for i in range(1, frequencies.shape[0]):
        assert not np.allclose(
            np.array([
                [[0.        ]], [[2.13215939]], [[0.        ]],
                [[2.16329116]], [[0.        ]], [[2.12899306]]]),
                init_energy_dir[..., i])


def test_collect_receiver_direct_sound(sample_walls):
    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        sample_walls, 1)

    position = np.array([0, 0, 0])
    view = np.array([1, 0, 0])
    up = np.array([0, 0, 1])
    path = os.path.join(
        os.path.dirname(__file__), 'test_data',
        'Genelec8020_DAF_2016_1x1.v17.ms.sofa')

    # create directivity
    directivity = sp.sound_object.DirectivityMS(path)
    sound_source = sp.sound_object.SoundSource(position, view, up, directivity)

    # set air attenuation
    frequencies = sf.read_sofa(path, verbose=False).N
    radiosity.set_air_attenuation(
        pf.FrequencyData(np.zeros_like(frequencies), frequencies))

    # try with directivity
    radiosity.init_source_energy(sound_source)
    radiosity.calculate_energy_exchange(343, 0.1, 1)

    receivers = pf.Coordinates([0.5, 0.5], 0, 0)
    direct_sound, delay_samples = radiosity.calculate_direct_sound(
        receivers)
    n_receivers = receivers.cshape[0]
    n_bins = radiosity.n_bins
    npt.assert_almost_equal(direct_sound.shape, (n_receivers, n_bins))
    npt.assert_almost_equal(delay_samples.shape, (n_receivers))

    npt.assert_almost_equal(direct_sound[0], direct_sound[1])
    npt.assert_almost_equal(delay_samples[0], delay_samples[1])


def test_collect_receiver_mono_direct_sound(sample_walls):
    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        sample_walls, 1)

    position = np.array([0, 0, 0])
    view = np.array([1, 0, 0])
    up = np.array([0, 0, 1])
    path = os.path.join(
        os.path.dirname(__file__), 'test_data',
        'Genelec8020_DAF_2016_1x1.v17.ms.sofa')

    # create directivity
    directivity = sp.sound_object.DirectivityMS(path)
    sound_source = sp.sound_object.SoundSource(position, view, up, directivity)

    # set air attenuation
    frequencies = sf.read_sofa(path, verbose=False).N
    radiosity.set_air_attenuation(
        pf.FrequencyData(np.zeros_like(frequencies), frequencies))

    # try with directivity
    radiosity.init_source_energy(sound_source)
    radiosity.calculate_energy_exchange(343, 0.001, .1)

    receivers = pf.Coordinates([0.5, 0.5], 0, 0)
    direct_sound, delay_samples = radiosity.calculate_direct_sound(
        receivers)

    etc = radiosity.collect_energy_receiver_mono(
        receivers, True)
    npt.assert_almost_equal(
        etc.time[
            np.arange(len(delay_samples)), : , delay_samples], direct_sound)


def test_collect_receiver_mono_direct_sound_with_brdf(
        sample_walls, sofa_data_diffuse_full_third_octave):
    brdf, sources, receivers = sofa_data_diffuse_full_third_octave
    radiosity = sp.DirectionalRadiosityFast.from_polygon(
        sample_walls, 1)

    position = np.array([0, 0, 0])
    view = np.array([1, 0, 0])
    up = np.array([0, 0, 1])
    path = os.path.join(
        os.path.dirname(__file__), 'test_data',
        'Genelec8020_DAF_2016_1x1.v17.ms.sofa')

    # create directivity
    directivity = sp.sound_object.DirectivityMS(path)
    sound_source = sp.sound_object.SoundSource(position, view, up, directivity)

    # set air attenuation
    radiosity.set_wall_brdf(
        np.arange(6), brdf, sources, receivers)

    # try with directivity
    radiosity.init_source_energy(sound_source)
    radiosity.calculate_energy_exchange(343, 0.001, .1)

    receivers = pf.Coordinates([0.5, 0.5], 0, 0)
    direct_sound, delay_samples = radiosity.calculate_direct_sound(
        receivers)

    etc = radiosity.collect_energy_receiver_mono(
        receivers, True)
    npt.assert_almost_equal(
        etc.time[
            np.arange(len(delay_samples)), : , delay_samples], direct_sound)




"""Test the radiosity module with directional Patches."""
import os

import numpy as np
import numpy.testing as npt
import pyfar as pf
import pytest
import sparrowpy.geometry as geo
import sparrowpy as sp
from sparrowpy.sound_object import Receiver, SoundSource


create_reference_files = False


@pytest.mark.parametrize('max_order_k', [2, 3])
def test_radiosity_directional_reference(max_order_k):
    """Test if the results changes."""
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data', 'ihta.E_sec_2.sofa')
    X = 10
    Y = 20
    Z = 15
    patch_size = 5
    ir_length_s = 5
    sampling_rate = 50

    ## create geometry
    ground = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, Y, 0], [0, Y, 0]], [1, 0, 0], [0, 0, 1])
    A_wall = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, 0, Z], [0, 0, Z]], [1, 0, 0], [0, 1, 0])
    B_wall = geo.Polygon(
        [[0, Y, 0], [X, Y, 0], [X, Y, Z], [0, Y, Z]], [1, 0, 0], [0, -1, 0])
    source = SoundSource([10, 6, 1], [0, 1, 0], [0, 0, 1])

    ## new approach
    radi = sp.DirectionalRadiosityKang(
        [ground, A_wall, B_wall], patch_size, max_order_k, ir_length_s,
        path_sofa, speed_of_sound=343, sampling_rate=sampling_rate)

    for patches in radi.patch_list:
        patches.directivity_data.freq = np.ones_like(
            patches.directivity_data.freq)

    radi.run(source)

    receiver = Receiver([20, 2, 1], [0, 1, 0], [0, 0, 1])
    irs_new = radi.energy_at_receiver(receiver)
    signal = pf.Signal(irs_new, sampling_rate)
    signal.time /= np.max(np.abs(signal.time))

    test_path = os.path.join(
        os.path.dirname(__file__), 'test_data')

    result = pf.io.read(
        os.path.join(
            test_path,
            f'simulation_X{X}_k{max_order_k}_{patch_size}m.far'))

    for i_frequency in range(signal.time.shape[0]):
        npt.assert_almost_equal(
            result['signal'].time[0, ...], signal.time[i_frequency, ...],
            decimal=4)
    npt.assert_almost_equal(result['signal'].times, signal.times)


@pytest.mark.parametrize('max_order_k', [2, 3])
def test_radiosity_directional_reference_read_write(max_order_k, tmpdir):
    """Test if the results changes."""
    X = 10
    Y = 20
    Z = 15
    patch_size = 5
    ir_length_s = 5
    sampling_rate = 50
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data', 'ihta.E_sec_2.sofa')

    ## create geometry
    ground = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, Y, 0], [0, Y, 0]], [1, 0, 0], [0, 0, 1])
    A_wall = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, 0, Z], [0, 0, Z]], [1, 0, 0], [0, 1, 0])
    B_wall = geo.Polygon(
        [[0, Y, 0], [X, Y, 0], [X, Y, Z], [0, Y, Z]], [1, 0, 0], [0, -1, 0])
    source = SoundSource([10, 6, 1], [0, 1, 0], [0, 0, 1])

    ## new approach
    radi = sp.DirectionalRadiosityKang(
        [ground, A_wall, B_wall], patch_size, max_order_k, ir_length_s,
        path_sofa, speed_of_sound=343, sampling_rate=sampling_rate)

    for patches in radi.patch_list:
        patches.directivity_data.freq = np.ones_like(
            patches.directivity_data.freq)

    radi.run(source)
    radi.write(os.path.join(tmpdir, 'radiosity.far'))
    radi = sp.DirectionalRadiosityKang.from_read(
        os.path.join(tmpdir, 'radiosity.far'))

    receiver = Receiver([20, 2, 1], [0, 1, 0], [0, 0, 1])
    irs_new = radi.energy_at_receiver(receiver)
    signal = pf.Signal(irs_new, sampling_rate)
    signal.time /= np.max(np.abs(signal.time))

    test_path = os.path.join(
        os.path.dirname(__file__), 'test_data')

    result = pf.io.read(
        os.path.join(
            test_path,
            f'simulation_X{X}_k{max_order_k}_{patch_size}m.far'))

    for i_frequency in range(signal.time.shape[0]):
        npt.assert_almost_equal(
            result['signal'].time[0, ...], signal.time[i_frequency, ...],
            decimal=4)
    npt.assert_almost_equal(result['signal'].times, signal.times)


@pytest.mark.parametrize('i_wall', [0, 1, 2, 3, 4, 5])
def test_init_energy_exchange(sample_walls, i_wall):
    """Test vs references for energy_exchange."""
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data', 'ihta.E_sec_2.sofa')
    patches = sp.PatchesDirectionalKang.from_sofa(
        sample_walls[i_wall], 0.2, [], 0, path_sofa)
    max_order_k = 3
    ir_length_s = 5
    source = SoundSource([0.5, 0.5, 0.5], [0, 1, 0], [0, 0, 1])
    patches.init_energy_exchange(
        max_order_k, ir_length_s, source, 1000, 346.18)
    normal = np.array(patches.normal)
    # check if the sources are in the right direction
    eps = 1e-10
    for i in range(3):
        if normal[i] > eps:
            assert all(patches.directivity_sources.cartesian[:, i] > -eps)
        if normal[i] < -eps:
            assert all(patches.directivity_sources.cartesian[:, i] < eps)
    # check if the receivers are in the right direction
    for i in range(3):
        if normal[i] > eps:
            assert all(patches.directivity_receivers.cartesian[:, i] > -eps)
        if normal[i] < -eps:
            assert all(patches.directivity_receivers.cartesian[:, i] < eps)


@pytest.mark.parametrize('patch_size', [
    0.5,
    1,
    ])
@pytest.mark.parametrize('i_wall', [0, 1, 2, 3, 4, 5])
def test_init_energy_exchange_directional_omni(
        sample_walls, patch_size, i_wall):
    """Test vs refernces for energy_exchange."""
    reference_path = os.path.join(
        os.path.dirname(__file__), 'test_data',
        f'reference_matrix_directional_patch_size{patch_size}.far')
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data', 'ihta.E_sec_2.sofa')
    patches = sp.PatchesDirectionalKang.from_sofa(
        sample_walls[i_wall], patch_size, [], 0, path_sofa)
    patches.directivity_data.freq = np.ones_like(
        patches.directivity_data.freq)
    max_order_k = 3
    ir_length_s = 5
    source = SoundSource([0.5, 0.5, 0.5], [0, 1, 0], [0, 0, 1])
    patches.init_energy_exchange(
        max_order_k, ir_length_s, source, 1000, 346.18)
    data = pf.io.read(reference_path)
    for i_frequency in range(patches.E_matrix.shape[0]):
        for i_rec in range(patches.E_matrix.shape[-1]):
            npt.assert_almost_equal(
                data['E_matrix'][0, ...],
                patches.E_matrix[i_frequency, ..., i_rec])


@pytest.mark.parametrize('perpendicular_walls', [
    [0, 2],
    [0, 4], [2, 0], [2, 4], [4, 0], [4, 2],
    [1, 2], [1, 3], [1, 4], [1, 5],
    [2, 1], [2, 5], [0, 3], [0, 5],
    [3, 0], [3, 1], [3, 4], [3, 5],
    [4, 1], [4, 3],
    [5, 0], [5, 1], [5, 2], [5, 3],
    ])
@pytest.mark.parametrize('patch_size', [
    0.5,
    1,
    ])
def test_directional_specular_reflections(
        sample_walls, perpendicular_walls, patch_size):
    """Test vs references for specular_reflections."""
    max_order_k=3
    ir_length_s=5
    sampling_rate = 1000
    speed_of_sound = 346.18
    wall_source = sample_walls[perpendicular_walls[0]]
    wall_receiver = sample_walls[perpendicular_walls[1]]
    path_reference = os.path.join(
        os.path.dirname(__file__), 'test_data',
        f'reference_specular_reflections_{patch_size}.far')
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data', 'specular_gaussian5.sofa')
    patch_1 = sp.PatchesDirectionalKang.from_sofa(
        wall_source, patch_size, [1], 0, path_sofa)
    patch_2 = sp.PatchesDirectionalKang.from_sofa(
        wall_receiver, patch_size, [0], 1, path_sofa)
    source = SoundSource([0.5, 0.5, 0.5], [0, 1, 0], [0, 0, 1])
    patches = [patch_1, patch_2]
    patch_1.calculate_form_factor(patches)
    patch_1.init_energy_exchange(
        max_order_k, ir_length_s, source, sampling_rate, speed_of_sound)
    patch_2.calculate_form_factor(patches)
    patch_2.init_energy_exchange(
        max_order_k, ir_length_s, source, sampling_rate, speed_of_sound)
    for k in range(1, max_order_k+1):
        patch_1.calculate_energy_exchange(
            patches, k, speed_of_sound, sampling_rate)
        patch_2.calculate_energy_exchange(
            patches, k, speed_of_sound, sampling_rate)
    receiver = Receiver([0.5, 0.5, 0.5], [0, 1, 0], [0, 0, 1])

    ir = 0
    ir += patch_1.energy_at_receiver(
        max_order_k, receiver,
        speed_of_sound=346.18, sampling_rate=1000)
    ir += patch_2.energy_at_receiver(
        max_order_k, receiver,
        speed_of_sound=346.18, sampling_rate=1000)

    if create_reference_files and (
            perpendicular_walls[0] == 0) and (perpendicular_walls[1] == 2):
        pf.io.write(path_reference, ir=ir)

    data = pf.io.read(path_reference)

    npt.assert_almost_equal(data['ir'], ir, decimal=4)


def test_PatchDirectional_to_from_dict(sample_walls):
    """Test if the results are correct with from_dict."""
    perpendicular_walls = [0, 2]
    patch_size = 0.2
    sampling_rate = 1000
    speed_of_sound = 346.18
    wall_source = sample_walls[perpendicular_walls[0]]
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data', 'specular_gaussian5.sofa')
    patch_1 = sp.PatchesDirectionalKang.from_sofa(
        wall_source, patch_size, [1], 0, path_sofa)
    patch_1.init_energy_exchange(
        1, 1, SoundSource([0.5, 0.5, 0.5], [0, 1, 0], [0, 0, 1]),
        sampling_rate, speed_of_sound)
    reconstructed_patch = sp.PatchesDirectionalKang.from_dict(
        patch_1.to_dict())
    assert reconstructed_patch.directivity_data == patch_1.directivity_data
    assert all(
        reconstructed_patch.directivity_sources ==
            patch_1.directivity_sources)
    assert all(
        reconstructed_patch.directivity_receivers
            == patch_1.directivity_receivers)
    npt.assert_array_equal(reconstructed_patch.E_matrix, patch_1.E_matrix)
    npt.assert_array_equal(
        reconstructed_patch.E_n_samples, patch_1.E_n_samples)
    npt.assert_array_equal(
        reconstructed_patch.form_factors, patch_1.form_factors)
    npt.assert_array_equal(reconstructed_patch.pts, patch_1.pts)
    npt.assert_array_equal(reconstructed_patch.up_vector, patch_1.up_vector)
    npt.assert_array_equal(reconstructed_patch.normal, patch_1.normal)
    assert reconstructed_patch.max_size == patch_1.max_size
    assert reconstructed_patch.other_wall_ids == patch_1.other_wall_ids
    assert reconstructed_patch.wall_id == patch_1.wall_id
    npt.assert_array_equal(reconstructed_patch.scattering, patch_1.scattering)
    npt.assert_array_equal(reconstructed_patch.absorption, patch_1.absorption)
    npt.assert_array_equal(
        reconstructed_patch.sound_attenuation_factor,
        patch_1.sound_attenuation_factor)


def test_RadiosityDirectional_to_from_dict():
    """Test if the results are correct with to_dict and from_dict."""
    max_order_k = 3
    patch_size = 5
    X = 10
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data', 'ihta.E_sec_2.sofa')
    Y = 20
    Z = 15
    ir_length_s = 5
    sampling_rate = 50

    # create geometry
    ground = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, Y, 0], [0, Y, 0]], [1, 0, 0], [0, 0, 1])
    A_wall = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, 0, Z], [0, 0, Z]], [1, 0, 0], [0, 1, 0])
    B_wall = geo.Polygon(
        [[0, Y, 0], [X, Y, 0], [X, Y, Z], [0, Y, Z]], [1, 0, 0], [0, -1, 0])
    source = SoundSource([10, 6, 1], [0, 1, 0], [0, 0, 1])

    # new approach
    radi = sp.DirectionalRadiosityKang(
        [ground, A_wall, B_wall], patch_size, max_order_k, ir_length_s,
        path_sofa, speed_of_sound=343, sampling_rate=sampling_rate)

    for patches in radi.patch_list:
        patches.directivity_data.freq = np.ones_like(
            patches.directivity_data.freq)

    radi.run(source)
    radi_dict = radi.to_dict()
    radi_reconstructed = sp.RadiosityKang.from_dict(radi_dict)

    # test
    assert isinstance(radi_reconstructed, sp.RadiosityKang)
    assert radi_reconstructed.patch_size == radi.patch_size
    assert radi_reconstructed.max_order_k == radi.max_order_k
    assert radi_reconstructed.ir_length_s == radi.ir_length_s
    assert radi_reconstructed.speed_of_sound == radi.speed_of_sound
    assert radi_reconstructed.sampling_rate == radi.sampling_rate
    assert len(radi_reconstructed.patch_list) == len(radi.patch_list)
    for patch, patch_reconstructed in zip(
            radi.patch_list, radi_reconstructed.patch_list, strict=True):
        np.testing.assert_array_equal(
            patch.pts, patch_reconstructed.pts)
        np.testing.assert_array_equal(
            patch.up_vector, patch_reconstructed.up_vector)
        np.testing.assert_array_equal(
            patch._normal, patch_reconstructed._normal)
        assert patch.max_size == patch_reconstructed.max_size
        assert patch.wall_id == patch_reconstructed.wall_id
        np.testing.assert_array_equal(
            patch.scattering, patch_reconstructed.scattering)
        np.testing.assert_array_equal(
            patch.absorption, patch_reconstructed.absorption)
        np.testing.assert_array_equal(
            patch.sound_attenuation_factor,
            patch_reconstructed.sound_attenuation_factor)
        np.testing.assert_array_equal(
            patch.E_matrix, patch_reconstructed.E_matrix)


def test_RadiosityDirectional_read_write(tmpdir):
    """Test if the results are correct with read and write."""
    max_order_k = 3
    patch_size = 5
    X = 10
    path_sofa = os.path.join(
        os.path.dirname(__file__), 'test_data', 'ihta.E_sec_2.sofa')
    Y = 20
    Z = 15
    ir_length_s = 5
    sampling_rate = 50

    # create geometry
    ground = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, Y, 0], [0, Y, 0]], [1, 0, 0], [0, 0, 1])
    A_wall = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, 0, Z], [0, 0, Z]], [1, 0, 0], [0, 1, 0])
    B_wall = geo.Polygon(
        [[0, Y, 0], [X, Y, 0], [X, Y, Z], [0, Y, Z]], [1, 0, 0], [0, -1, 0])
    source = SoundSource([10, 6, 1], [0, 1, 0], [0, 0, 1])

    # new approach
    radi = sp.DirectionalRadiosityKang(
        [ground, A_wall, B_wall], patch_size, max_order_k, ir_length_s,
        path_sofa, speed_of_sound=343, sampling_rate=sampling_rate)

    for patches in radi.patch_list:
        patches.directivity_data.freq = np.ones_like(
            patches.directivity_data.freq)

    radi.run(source)
    json_path = os.path.join(tmpdir, 'radi.json')
    radi.write(json_path)
    radi_reconstructed = sp.DirectionalRadiosityKang.from_read(
        json_path)

    # test
    assert isinstance(radi_reconstructed, sp.DirectionalRadiosityKang)
    assert radi_reconstructed.patch_size == radi.patch_size
    assert radi_reconstructed.max_order_k == radi.max_order_k
    assert radi_reconstructed.ir_length_s == radi.ir_length_s
    assert radi_reconstructed.speed_of_sound == radi.speed_of_sound
    assert radi_reconstructed.sampling_rate == radi.sampling_rate
    assert len(radi_reconstructed.patch_list) == len(radi.patch_list)
    for patch, patch_reconstructed in zip(
            radi.patch_list, radi_reconstructed.patch_list, strict=True):
        np.testing.assert_array_equal(
            patch.pts, patch_reconstructed.pts)
        np.testing.assert_array_equal(
            patch.up_vector, patch_reconstructed.up_vector)
        np.testing.assert_array_equal(
            patch._normal, patch_reconstructed._normal)
        assert patch.max_size == patch_reconstructed.max_size
        assert patch.wall_id == patch_reconstructed.wall_id
        np.testing.assert_array_equal(
            patch.scattering, patch_reconstructed.scattering)
        np.testing.assert_array_equal(
            patch.absorption, patch_reconstructed.absorption)
        np.testing.assert_array_equal(
            patch.sound_attenuation_factor,
            patch_reconstructed.sound_attenuation_factor)
        np.testing.assert_array_equal(
            patch.E_matrix, patch_reconstructed.E_matrix)




import numpy.testing as npt
import pytest
import numpy as np
import pyfar as pf
import sparrowpy as sp
bpy = pytest.importorskip("bpy")
import sparrowpy.utils.blender as bh  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402

@pytest.mark.parametrize("origin", [np.array([0.,1.,3.])])
@pytest.mark.parametrize("point", [np.array([0.,1.,-1])])
@pytest.mark.parametrize("plpt", [np.array([1.,1.,0.])])
@pytest.mark.parametrize("pln", [np.array([0.,0.,1.])])
@pytest.mark.parametrize("solution", [np.array([0.,1.,0.])])
def test_point_plane_projection(origin: np.ndarray, point: np.ndarray,
                                plpt: np.ndarray, pln: np.ndarray, solution):
    """Ensure correct projection of rays into plane."""
    out = sp.geometry._project_to_plane(origin, point, plpt, pln)

    npt.assert_array_equal(solution,out)



@pytest.mark.parametrize("point", [
    np.array([0.,0.,0.]),
    np.array([0.,2.,0.]),
    ])
@pytest.mark.parametrize("plpt", [
    np.array([[1.,1.,0.],[-1.,1.,0.],[-1.,-1.,0.],[1.,-1.,0.]]),
    ])
@pytest.mark.parametrize("pln", [np.array([0.,0.,1.])])
def test_point_in_polygon(point, plpt, pln):
    """Ensure correct projection of rays into plane."""
    out = sp.geometry._point_in_polygon(point3d=point,
                                 polygon3d=plpt,
                                 plane_normal=pln)

    if abs(point[0]) > 1. or abs(point[1]) > 1:
        solution = False
    else:
        solution = True

    assert solution==out


@pytest.mark.parametrize("point", [
    np.array([0.,0.,.5]),
    np.array([0.,0.,0.]),
    np.array([0.,0.,2.]),
    np.array([0.,0.,-1.]),
    np.array([0.,3.,-1.]),
    ])
@pytest.mark.parametrize("origin", [np.array([0.,0.,1.])])
@pytest.mark.parametrize("plpt", [
    3*np.array([[1.,1.,0.],[-1.,1.,0.],[-1.,-1.,0.],[1.,-1.,0.]]),
    3*np.array([[1.,1.,-1.],[-1.,1.,-1.],[-1.,-1.,1.],[1.,-1.,1.]]),
    ])

def test_basic_visibility(point, origin, plpt):
    """Test basic_visibility function."""

    pln = np.cross(plpt[1]-plpt[0],plpt[2]-plpt[1])
    pln /= np.linalg.norm(pln)

    out = sp.geometry._basic_visibility(eval_point=point,vis_point=origin,
                              surf_points=plpt,surf_normal=pln)

    if (np.dot(point-origin,pln)>0 or
        point[2]>=0 ):
        solution = 1
    else:
        solution = 0
    assert solution==out

@pytest.mark.parametrize("model", [
    "./tests/test_data/cube_simple.blend",
    "./tests/test_data/cube.blend",
    "./tests/test_data/cube_blocked.blend",
    ])
def test_vis_matrix_assembly(model):
    """Check if visibility matrices are correctly assembled."""
    m1,m2 = bh.read_geometry_file(model)

    patches_points = np.empty((len(m1["conn"]),len(m1["conn"][0]),3))
    patches_centers = np.empty((len(m1["conn"]),3))
    patches_normals = np.empty_like(patches_centers)

    for i in range(len(m1["conn"])):
            patches_points[i]=m1["verts"][m1["conn"][i]]
            patches_centers[i]=sp.geometry._calculate_center(m1["verts"][m1["conn"][i]])
            patches_normals[i]=np.cross(m1["verts"][m1["conn"][i]][1]
                                            -m1["verts"][m1["conn"][i]][0],
                                      m1["verts"][m1["conn"][i]][2]
                                            -m1["verts"][m1["conn"][i]][0])
            patches_normals[i]/=np.linalg.norm(patches_normals[i])

    for m in [m1,m2]:
        surfs=m

        surfs_points = []
        surfs_normals =list([np.empty((3,))]*len(surfs["conn"]))

        if model=="./tests/test_data/cube.blend":
            solution=np.zeros((patches_centers.shape[0],patches_centers.shape[0]),
                           dtype=bool)
            for i in range(solution.shape[0]):
                for j in range(i+1,solution.shape[1]):
                    ray=patches_centers[j]-patches_centers[i]
                    ray/=np.linalg.norm(ray)
                    if np.dot(patches_normals[i], ray)>1e-6:
                        solution[i,j]=True
        elif model=="./tests/test_data/cube_blocked.blend":
            solution=np.zeros((patches_centers.shape[0],patches_centers.shape[0]),
                            dtype=bool)
            for ids in A:
                solution[ids[0],ids[1]]=True
        elif model=="./tests/test_data/cube_simple.blend":
            solution=np.zeros((patches_centers.shape[0],patches_centers.shape[0]),
                            dtype=bool)
            for i in range(solution.shape[0]):
                for j in range(i+1,solution.shape[1]):
                        solution[i,j]=True

        for i in range(len(surfs["conn"])):
            surfs_points.append(surfs["verts"][surfs["conn"][i]])
            surfs_normals[i]=np.cross(surfs["verts"][m["conn"][i]][1]
                                            -m["verts"][m["conn"][i]][0],
                                      m["verts"][m["conn"][i]][2]
                                            -m["verts"][m["conn"][i]][0])
            surfs_normals[i]/=np.linalg.norm(surfs_normals[i])

        vis_matrix = sp.geometry._check_patch2patch_visibility(
                                            patches_center=patches_centers,
                                           surf_normal=surfs_normals,
                                           surf_points=surfs_points)

        plt.imsave(model[:-6]+"_vis.pdf",vis_matrix)
        plt.imsave(model[:-6]+"_sol.pdf",solution)
        npt.assert_array_equal(vis_matrix,solution)

A=[
    [0,1],
    [0,3],
    [0,4],
    [0,5],
    [0,6],
    [0,9],
    [0,14],
    [1,3],
    [1,4],
    [1,6],
    [1,7],
    [1,9],
    [1,14],
    [2,8],
    [2,9],
    [2,10],
    [2,11],
    [2,12],
    [2,13],
    [2,15],
    [3,5],
    [3,6],
    [3,7],
    [3,8],
    [3,9],
    [3,10],
    [3,11],
    [3,12],
    [3,13],
    [3,14],
    [3,15],
    [4,5],
    [4,6],
    [4,7],
    [4,9],
    [4,14],
    [5,6],
    [5,7],
    [5,9],
    [5,14],
    [6,7],
    [6,9],
    [6,14],
    [7,9],
    [7,14],
    [8,9],
    [8,10],
    [8,11],
    [8,12],
    [8,15],
    [9,10],
    [9,12],
    [9,13],
    [9,15],
    [10,11],
    [10,12],
    [10,13],
    [10,15],
    [11,12],
    [11,13],
    [11,15],
    [12,13],
    [13,15],
    [11,12],
   ]


def test_source_vis(basicscene):
    """Test visibility check between source and patches."""

    radi=sp.DirectionalRadiosityFast.from_polygon(basicscene["walls"],
                                                  patch_size=1.)

    radi.init_source_energy(pf.Coordinates(3.,3.,3.))
    npt.assert_equal(radi._energy_init_source,
                     np.zeros_like(radi._energy_init_source))
    npt.assert_equal(radi._source_visibility,
                     np.zeros_like(radi._source_visibility))

    radi.init_source_energy(pf.Coordinates(.5, .5, .5))
    npt.assert_equal(radi._source_visibility,
                     np.ones_like(radi._source_visibility))
    assert (radi._energy_init_source != 0).any()


def test_receiver_vis(basicscene):
    """Test visibility check between source and patches."""

    radi=sp.DirectionalRadiosityFast.from_polygon(basicscene["walls"],
                                                  patch_size=1.)

    radi.init_source_energy(pf.Coordinates(.5,.5,.5))

    radi.calculate_energy_exchange(speed_of_sound=343,
                                   etc_time_resolution=.2,
                                   etc_duration=1.)

    etc = radi.collect_energy_receiver_mono(pf.Coordinates([3.,-5.,.5],  #x
                                                           [3., 5.,.5],  #y
                                                           [3., 4.,.5])) #z

    npt.assert_equal(etc.time[0:2,0,:],
                     np.array([np.zeros((5)),
                              np.zeros((5))]),
                     )
    assert (etc.time[-1,0,:] != 0).any()



import sofar as sf
import numpy.testing as npt
import numpy as np


def test_brdf_s_0(brdf_s_0):
    sofa = sf.read_sofa(brdf_s_0)
    for i in range(4):
        npt.assert_almost_equal(sofa.Data_Real[i - 2, i], 1.1026578)
        npt.assert_almost_equal(
            np.sum(sofa.Data_Real[i - 2]), 1.1026578)


def test_brdf_s_1(brdf_s_1):
    sofa = sf.read_sofa(brdf_s_1)
    npt.assert_almost_equal(sofa.Data_Real, 1 / np.pi)



"""Module for the radiosity simulation."""
import numpy as np
import deepdiff
import pyfar as pf
import sparrowpy.form_factor.universal as form_factor
from sparrowpy import ( geometry, sound_object )
try:
    import numba
    prange = numba.prange
except ImportError:
    numba = None
    prange = range


class DirectionalRadiosityFast():
    """Radiosity object for directional scattering coefficients."""

    _walls_points: np.ndarray
    _walls_normal: np.ndarray
    _walls_up_vector: np.ndarray
    _patches_points: np.ndarray
    _n_patches: int
    _patch_to_wall_ids: np.ndarray

    # geometrical data
    _visibility_matrix: np.ndarray
    _visible_patches: np.ndarray
    _form_factors: np.ndarray
    _form_factors_tilde: np.ndarray

    # general data for material and medium data
    _frequencies: np.ndarray
    _brdf: np.ndarray
    _brdf_index: np.ndarray
    _brdf_incoming_directions: list[pf.Coordinates]
    _brdf_outgoing_directions: list[pf.Coordinates]
    _patch_2_brdf_outgoing_index: np.ndarray

    _air_attenuation: np.ndarray
    _speed_of_sound: float

    # etc metadata
    _etc_time_resolution: float
    _etc_duration: float

    # etc results
    _distance_patches_to_source: np.ndarray
    _energy_init_source: np.ndarray
    _energy_exchange_etc: np.ndarray


    def __init__(
            self,
            walls_points:np.ndarray,
            walls_normal:np.ndarray,
            walls_up_vector:np.ndarray,
            patches_points:np.ndarray,
            n_patches:int,
            patch_to_wall_ids:np.ndarray,
            visibility_matrix:np.ndarray=None,
            visible_patches:np.ndarray=None,
            form_factors:np.ndarray=None,
            form_factors_tilde:np.ndarray=None,
            frequencies:np.ndarray=None,
            brdf:list[np.ndarray]=None,
            brdf_index:np.ndarray=None,
            brdf_incoming_directions:list[pf.Coordinates]=None,
            brdf_outgoing_directions:list[pf.Coordinates]=None,
            patch_2_brdf_outgoing_index:np.ndarray=None,
            air_attenuation:np.ndarray=None,
            speed_of_sound:float=None,
            etc_time_resolution:float=None,
            etc_duration:float=None,
            distance_patches_to_source:np.ndarray=None,
            energy_init_source:np.ndarray=None,
            energy_exchange_etc:np.ndarray=None):
        """Create a Radiosity object for directional implementation.

        Parameters
        ----------
        walls_points : np.ndarray
            edge points of all walls in cartesian of shape
            (n_walls, n_points, 3)
        walls_normal : np.ndarray
            normals of all walls of shape (n_walls, 3)
        walls_up_vector : np.ndarray
            uo vector of all walls of shape (n_walls, 3)
        patches_points : np.ndarray
            edge points of all patches in cartesian of shape
            (n_patches, n_points, 3)
        n_patches : int
            number of patches
        patch_to_wall_ids : np.ndarray
            maps each patch to a wall of shape (n_patches)
        visibility_matrix : np.ndarray, optional
            patch to patch boolean visibility matrix, by default None.
        visible_patches : np.ndarray, optional
            list of all patches which are visible, by default None
        form_factors : np.ndarray, optional
            geometrical form factor from each patch to each other patch,
            by default None.
        form_factors_tilde : np.ndarray, optional
            Form factor including air attenuation and BRDF of shape
            (n_patches, n_patches, n_outgoing_directions, n_bins), by
            default None
        frequencies : np.ndarray, optional
            frequency vector used for the simulation. , by default None
        brdf : list[np.ndarray], optional
            brdf in its raw postprocessed data format, by default None
        brdf_index : np.ndarray, optional
            maps brdfs to walls, must be of shape (n_walls, ), by default None
        brdf_incoming_directions : list[pf.Coordinates], optional
            incoming direction of brdfs per wall, by default None
        brdf_outgoing_directions : list[pf.Coordinates], optional
            outgoing directions of brdfs per wall, by default None
        patch_2_brdf_outgoing_index: np.ndarray
            map of patch positions to relative scattering directions indices
        air_attenuation : np.ndarray, optional
            air attenuation coefficients for each frequency, needs to be of
            shape (n_bins), by default None
        speed_of_sound : float, optional
            speed of sound in m/s, by default None
        etc_time_resolution : float, optional
            time resolution fo the etc, by default None
        etc_duration : float, optional
            duration fo the etc in seconds, by default None
        distance_patches_to_source : np.ndarray, optional
            distance from the source to each patch, need to be of shape
            (n_patches), by default None
        energy_init_source : np.ndarray, optional
            initial energy from source to patches,
            need to be of shape (n_patches, n_outgoing_directions, n_bins),
            by default None
        energy_exchange_etc : np.ndarray, optional
            etc of the energy exchange, must be of shape
            (n_patches, n_outgoing_directions, n_bins, n_samples),
            by default None
        """
        # convert inputs
        walls_points = np.atleast_3d(walls_points)
        walls_up_vector = np.atleast_2d(walls_up_vector)
        walls_normal = np.atleast_2d(walls_normal)
        patches_points = np.atleast_3d(patches_points)
        patch_to_wall_ids = np.atleast_1d(np.array(
            patch_to_wall_ids, dtype=int))
        if frequencies is not None:
            frequencies = np.array(frequencies)
        if visible_patches is not None:
            visible_patches = np.array(visible_patches)
        if visibility_matrix is not None:
            visibility_matrix = np.array(visibility_matrix)
        if form_factors is not None:
            form_factors = np.array(form_factors)
        if form_factors_tilde is not None:
            form_factors_tilde = np.array(form_factors_tilde)
        if patch_2_brdf_outgoing_index is not None:
            patch_2_brdf_outgoing_index = np.array(patch_2_brdf_outgoing_index,
                                                   dtype=np.int64)
        if brdf is not None:
            brdf = [np.array(b) for b in brdf]
        if air_attenuation is not None:
            air_attenuation = np.array(air_attenuation)
        if speed_of_sound is not None:
            speed_of_sound = float(speed_of_sound)
        if etc_time_resolution is not None:
            etc_time_resolution = float(etc_time_resolution)
        if etc_duration is not None:
            etc_duration = float(etc_duration)
        if distance_patches_to_source is not None:
            distance_patches_to_source = np.array(distance_patches_to_source)
        if energy_init_source is not None:
            energy_init_source = np.array(energy_init_source)
        if energy_exchange_etc is not None:
            energy_exchange_etc = np.array(energy_exchange_etc)

        self._walls_points = walls_points
        self._walls_normal = walls_normal
        self._walls_up_vector = walls_up_vector
        self._patches_points = patches_points
        self._n_patches = n_patches
        self._patch_to_wall_ids = patch_to_wall_ids

        # geometrical data
        self._visibility_matrix = visibility_matrix
        self._visible_patches = visible_patches
        self._form_factors = form_factors
        self._form_factors_tilde = form_factors_tilde

        # general data for material and medium data
        self._frequencies = frequencies
        self._brdf = brdf
        self._brdf_index = brdf_index
        self._brdf_incoming_directions = brdf_incoming_directions
        self._brdf_outgoing_directions = brdf_outgoing_directions
        self._patch_2_brdf_outgoing_index = patch_2_brdf_outgoing_index

        self._air_attenuation = air_attenuation
        self._speed_of_sound = speed_of_sound

        # etc metadata
        self._etc_time_resolution = etc_time_resolution
        self._etc_duration = etc_duration

        # etc results
        self._distance_patches_to_source = distance_patches_to_source
        self._energy_init_source = energy_init_source
        self._energy_exchange_etc = energy_exchange_etc

        self.check()

    def check(self):
        """Check the input data for consistency."""

        n_walls = self._walls_points.shape[0]
        if (self._walls_points.shape[0] != n_walls) or \
                (self._walls_points.shape[2] != 3):
            raise ValueError(
                "Walls need to be of shape (n_walls, n_points, 3)")
        if self._walls_up_vector.shape != (n_walls, 3):
            raise ValueError(
                "Up vector of walls need to be of shape (n_walls, 3)")
        if self._walls_normal.shape != (n_walls, 3):
            raise ValueError(
                "Normal of walls need to be of shape (n_walls, 3)")

        # input checks Patches
        if (self._patches_points.shape[0] != self.n_patches) or \
                (self._patches_points.shape[2] != 3):
            raise ValueError(
                "Patches need to be of shape (n_patches, n_points, 3)")
        if self._patch_to_wall_ids.shape != (self.n_patches,):
            raise ValueError(
                "patch_to_wall_ids need to be of shape (n_patches,)")
        if any(i not in np.arange(
                n_walls) for i in set(self._patch_to_wall_ids)):
            raise ValueError(
                "patch_to_wall_ids does contain other ids than range(n_walls)")
        if any(i not in set(
                self._patch_to_wall_ids) for i in np.arange(n_walls)):
            raise ValueError(
                "patch_to_wall_ids does contain other ids than range(n_walls)")

        # check frequencies
        n_bins = 1
        if self._frequencies is not None:
            if len(self._frequencies.shape) != 1:
                raise ValueError(
                    "Frequencies need to be of shape (n_bins,)")
            n_bins = self._frequencies.size

        # check form factors
        if self._form_factors is not None:
            if self._form_factors.shape != (self.n_patches, self.n_patches):
                raise ValueError(
                    "form_factors need to be of shape (n_patches, n_patches)")

        # check brdf
        n_outgoing_directions = 1
        if self._brdf_index is not None:
            if len(self._brdf_index) != n_walls:
                raise ValueError(
                    "brdf_index need to be of shape (n_walls,)")
        if self._brdf_incoming_directions is not None:
            if any(not isinstance(i, pf.Coordinates) \
                   for i in self._brdf_incoming_directions):
                raise ValueError(
                    "brdf_incoming_directions need to be a list of type "
                    "pf.Coordinates")
        if self._brdf_outgoing_directions is not None:
            if any(not isinstance(i, pf.Coordinates) \
                   for i in self._brdf_outgoing_directions):
                raise ValueError(
                    "brdf_outgoing_directions need to be a list of type "
                    "pf.Coordinates")
            n_outgoing_directions = self._brdf_outgoing_directions[0].csize

        # check form_factors_tilde
        if self._form_factors_tilde is not None:
            if self._form_factors_tilde.shape != (
                    self.n_patches, self.n_patches,
                    n_outgoing_directions, n_bins):
                raise ValueError(
                    "form_factors_tilde need to be of shape "
                    "(n_patches, n_patches, n_outgoing_directions, n_bins)")

        # check air_attenuation
        if self._air_attenuation is not None:
            if len(self._air_attenuation.shape) != 1:
                raise ValueError(
                    "Air attenuation need to be of shape (n_bins,)")
            if self._air_attenuation.shape[0] != n_bins:
                raise ValueError(
                    "Air attenuation need to be of shape (n_bins,)")

        # check speed_of_sound
        if self._speed_of_sound is not None:
            if self._speed_of_sound <= 0:
                raise ValueError(
                    "Speed of sound must be positive and non-zero")

        # check etc_time_resolution
        if self._etc_time_resolution is not None:
            if self._etc_time_resolution <= 0:
                raise ValueError(
                    "Time resolution must be positive and non-zero")

        # check etc_duration
        if self._etc_duration is not None:
            if self._etc_duration <= 0:
                raise ValueError(
                    "Duration must be positive and non-zero")

        # check distance_patches_to_source
        if self._distance_patches_to_source is not None:
            if self._distance_patches_to_source.shape != (self.n_patches,):
                raise ValueError(
                    "distance_patches_to_source need to be of shape "
                    "(n_patches,)")

        # check energy_init_source
        if self._energy_init_source is not None:
            if self._energy_init_source.shape != (
                    self.n_patches, n_outgoing_directions, n_bins):
                raise ValueError(
                    "energy_init_source need to be of shape "
                    "(n_patches, n_outgoing_directions, n_bins)")

        # check energy_exchange_etc
        if self._energy_exchange_etc is not None:
            n_samples = int(
                self._etc_duration/self._etc_time_resolution)
            if self._energy_exchange_etc.shape != (
                    self.n_patches,
                    n_outgoing_directions, n_bins, n_samples):
                raise ValueError(
                    "energy_exchange_etc need to be of shape "
                    "(n_patches, n_outgoing_directions, n_bins, n_samples)")


    @classmethod
    def from_polygon(
            cls, polygon_list, patch_size):
        """Create a Radiosity object for directional scattering coefficients.

        Parameters
        ----------
        polygon_list : list[Polygon]
            list of patches
        patch_size : float
            maximal patch size in meters.

        """
        # save wall information
        walls_points = np.array([p.pts for p in polygon_list])
        walls_normal = np.array([p.normal for p in polygon_list])
        walls_up_vector = np.array([p.up_vector for p in polygon_list])

        # create patches
        (
            patches_points, _,
            n_patches, patch_to_wall_ids) = geometry._process_patches(
            walls_points, walls_normal, patch_size, len(polygon_list))
        # create radiosity object
        return cls(
            walls_points, walls_normal, walls_up_vector,
            patches_points, n_patches,
            patch_to_wall_ids)

    def bake_geometry(self):
        """Bake the geometry by calculating all the form factors.

        """
        # Check the visibility between patches.
        self._visibility_matrix = geometry._check_patch2patch_visibility(
            self.patches_center, self.patches_normal, self.patches_points)

        n_combinations = np.sum(self.visibility_matrix)
        visible_patches = np.empty((n_combinations, 2), dtype=np.int32)
        i_counter = 0
        for i_source in range(self.n_patches):
            for i_receiver in range(self.n_patches):
                if not self.visibility_matrix[i_source, i_receiver]:
                    continue
                visible_patches[i_counter, 0] = i_source
                visible_patches[i_counter, 1] = i_receiver
                i_counter += 1
        self._visible_patches = visible_patches

        self._form_factors = form_factor.patch2patch_ff_universal(
            self.patches_points, self.patches_normal,
            self.patches_area, self._visible_patches)

        # Calculate the form factors with directivity.
        if self._brdf_incoming_directions is not None:
            sources_array = np.array(
                [s.cartesian for s in self._brdf_incoming_directions])
            receivers_array = np.array(
                [s.cartesian for s in self._brdf_outgoing_directions])
            scattering_index = np.array(self._brdf_index)
            scattering = np.array(self._brdf)

            # preload patch_2_brdf_outgoing_index map with invalid entries
            self._patch_2_brdf_outgoing_index = (
                        receivers_array.shape[1] *
                                np.ones((self.n_patches,self.n_patches),dtype=np.int64))

            for j in range(self.n_patches):
                vis = np.where(
                    (self.visibility_matrix+self.visibility_matrix.T)[:,j])
                self._patch_2_brdf_outgoing_index[vis,j]=get_scattering_data_receiver_index(
                        pos_i=self.patches_center[vis],pos_j=self.patches_center[j],
                        receivers=receivers_array,
                        wall_id_i=self._patch_to_wall_ids[vis],
                    )
        else:
            sources_array = None
            receivers_array = None
            scattering_index = None
            scattering = None
            self._patch_2_brdf_outgoing_index = np.zeros(
                                        (self.n_patches,self.n_patches),
                                        dtype=np.int64)

        n_bins = 1 if self._frequencies is None else self.n_bins

        self._form_factors_tilde = \
            _form_factors_with_directivity_dim(
            self.visibility_matrix, self.form_factors, n_bins,
            self.patches_center, self.patches_area,
            self._air_attenuation,
            self._patch_to_wall_ids, scattering,
            scattering_index,
            sources_array, receivers_array)


    def init_source_energy(
            self, source):
        """Initialize the source energy.

        Parameters
        ----------
        source : pf.Coordinates, sparrowpy.sound_object.SoundSource
            definition of the source position for Coordinates object and
            orientation and directivity for SoundSource object. If no
            directivity is given, the directivity is set to 1 for all
            frequencies.

        """
        if isinstance(source, pf.Coordinates):
            if source.cshape != (1, ):
                raise ValueError('just one source position is allowed.')
            source_position = source.cartesian[0]
        elif isinstance(source, sound_object.SoundSource):
            source_position = source.position
        self._source = source


        patch_to_wall_ids = self._patch_to_wall_ids
        if self._brdf_incoming_directions is None:
            frequencies = np.array([0]) if self._frequencies is None else \
                self._frequencies
            self.set_wall_brdf(
                np.arange(self.n_walls),
                pf.FrequencyData(np.ones_like(frequencies), frequencies),
                pf.Coordinates(0, 0, 1, weights=1),
                pf.Coordinates(0, 0, 1, weights=1))
            self._frequencies = frequencies
        if self._air_attenuation is None:
            frequencies = np.array([0]) if self._frequencies is None else \
                self._frequencies
            self.set_air_attenuation(
                pf.FrequencyData(np.zeros_like(frequencies), frequencies))
            self._frequencies = frequencies
        n_bins = self.n_bins
        vi = np.array(
            [s.cartesian for s in self._brdf_incoming_directions])
        vo = np.array(
            [s.cartesian for s in self._brdf_outgoing_directions])
        brdf = np.array(self._brdf)
        brdf_index = self._brdf_index
        patches_center = self.patches_center
        source_visibility = geometry._check_point2patch_visibility(
                                        eval_point=source_position,
                                        patches_center=patches_center,
                                        surf_points=self.walls_points,
                                        surf_normal=self.walls_normal)
        self._source_visibility = source_visibility
        energy_0, distance_0 = form_factor._source2patch_energy_universal(
            source_position, patches_center, self.patches_points,
            source_visibility,
            self._air_attenuation, n_bins)

        # of shape (n_patches, n_directions, n_bins)
        energy_0_dir = _add_directional(
            energy_0, source_position,
            patches_center, n_bins, patch_to_wall_ids,
            vi, vo, brdf, brdf_index)

        # add directivity if given
        if isinstance(source, sound_object.SoundSource):
            n_patches = patches_center.shape[0]
            n_directions = vo.shape[1]

            if source.directivity is not None:
                directivity = np.zeros((n_patches, n_directions, n_bins))
                for i_frequency in range(n_bins):
                    directivity_local = np.real(source.get_directivity(
                        patches_center, self._frequencies[i_frequency]))
                    if n_directions == 1:
                        directivity[:, :, i_frequency] = directivity_local[
                            :, np.newaxis]
                    else:
                        directivity[:, :, i_frequency] = np.repeat(
                            directivity_local[..., np.newaxis],
                            n_directions,
                            axis=-1)
                energy_0_dir *= directivity
            else:
                energy_0_dir *= 1

        self._energy_init_source = energy_0_dir
        self._distance_patches_to_source = distance_0

    def calculate_energy_exchange(
            self, speed_of_sound,
            etc_time_resolution,
            etc_duration,
            max_reflection_order=-1,
            recalculate=False):
        """Calculate the energy exchange between patches.
        # todo? make first order to first reflection.
        """
        n_samples = int(etc_duration/etc_time_resolution)

        patches_center = self.patches_center
        distance_0 = self._distance_patches_to_source
        n_patches = self.n_patches
        distance_i_j = np.empty((n_patches, n_patches))

        for i in range(n_patches):
            for j in range(n_patches):
                distance_i_j[i, j] = np.linalg.norm(
                    patches_center[i, :]-patches_center[j, :])

        energy_0_dir = self._energy_init_source

        if self._energy_exchange_etc is None or recalculate:
            # energy exchange etc
            # of shape (n_patches, n_directions, n_bins, n_samples)
            if max_reflection_order < 1:
                self._energy_exchange_etc = \
                    _energy_exchange_init_energy(
                        n_samples, energy_0_dir, distance_0,
                        speed_of_sound, etc_time_resolution,
                        )
            else:

                self._energy_exchange_etc = _energy_exchange(
                    n_samples, energy_0_dir, distance_0, distance_i_j,
                    self._form_factors_tilde,
                    self._patch_2_brdf_outgoing_index,
                    speed_of_sound, etc_time_resolution,
                    max_reflection_order,
                    self._visible_patches)

        self._etc_time_resolution = float(etc_time_resolution)
        self._speed_of_sound = float(speed_of_sound)
        self._etc_duration = float(etc_duration)

    def collect_energy_receiver_mono(self, receivers, direct_sound=False):
        """Collect the energy at the receivers.

        Parameters
        ----------
        receivers : pf.Coordinates
            receiver Coordinates in of cshape (n_receivers).
        direct_sound : bool, optional
            If True, the direct sound is collected as well, by default False.
            The direct sound includes spreading loss, air attenuation and
            source directivity.

        Returns
        -------
        etc : pf.TimeData
            energy collected at the receiver in cshape
            (n_receivers, n_bins)
        """
        if not isinstance(direct_sound, bool):
            raise ValueError(
                "direct_sound must be of type boolean")
        etc = self.collect_energy_receiver_patchwise(receivers)
        etc.time = np.sum(etc.time, axis=1)

        if direct_sound:
            direct_sound, n_sample_delay = self.calculate_direct_sound(
                receivers)

            # add the direct sound to the etc
            i_receivers = np.arange(len(n_sample_delay))
            etc.time[i_receivers, :, n_sample_delay] += direct_sound

        return etc


    def calculate_direct_sound(self, receivers):
        """Calculate the direct sound at the receivers.

        It includes the spreading loss, air attenuation and
        source directivity.

        Parameters
        ----------
        receivers : pf.Coordinates
            receiver Coordinates in of cshape (n_receivers).

        Returns
        -------
        direct_sound : np.ndarray
            energy of the direct sound at the receivers in shape
            (n_receivers, n_bins)
        n_sample_delay : np.ndarray
            number of samples for the delay at the receivers in shape
            (n_receivers, )
        """
        if not isinstance(receivers, pf.Coordinates):
            raise ValueError(
                "Receiver positions must be of type pf.Coordinates")

        # calculate distance from source to receivers
        if isinstance(self._source, pf.Coordinates):
            source_position = self._source
        else:
            source_position = pf.Coordinates(*self._source.position)
        r = (receivers-source_position).radius

        # calculate spreading loss for direct sound
        direct_sound = np.ones(
            (receivers.cshape[0], self.n_bins), dtype=float)
        direct_sound *= (1/(4 * np.pi * r**2))[:, np.newaxis]

        # add air attenuation
        if self._air_attenuation is not None:
            for i in range(self.n_bins):
                direct_sound[:, i] *= np.exp(
                    -self._air_attenuation[i] * r)

        # add source directivity
        if isinstance(self._source, sound_object.SoundSource):
            for i in range(self.n_bins):
                direct_sound[:, i] *= np.real(self._source.get_directivity(
                    np.squeeze(receivers.cartesian), self._frequencies[i]))

        # calculate the number of samples for the delay
        n_sample_delay = np.array(
            r/self.speed_of_sound/self._etc_time_resolution, dtype=int)

        return direct_sound, n_sample_delay


    def collect_energy_receiver_patchwise(self, receivers):
        """Collect the energy for each patch at the receivers without summing
        up the patches.

        Parameters
        ----------
        receivers : pf.Coordinates
            receiver Coordinates in of cshape (n_receivers).

        Returns
        -------
        etc : pf.TimeData
            energy collected at the receiver in cshape
            (n_receivers, n_patches, n_bins)
        """
        if not isinstance(receivers, pf.Coordinates):
            raise ValueError(
                "Receiver positions must be of type pf.Coordinates")
        if receivers.cdim != 1:
            raise ValueError(
                "Receiver positions must be of shape (n_receivers, 3)")
        etc_data = self._collect_energy_patches(
            receivers.cartesian, propagation_fx=True)
        times = np.arange(etc_data.shape[-1]) * self._etc_time_resolution
        return pf.TimeData(etc_data, times)

    def _collect_energy_patches(
            self, receiver_pos,
            propagation_fx=False):
        """Collect patch histograms as detected by receiver."""
        air_attenuation = self._air_attenuation
        patches_points = self._patches_points
        n_patches = self.n_patches
        n_bins = self.n_bins

        receiver_pos = np.atleast_2d(receiver_pos)

        n_receivers = receiver_pos.shape[0]

        patches_center = self.patches_center
        patches_receiver_distance = np.empty(
            [n_receivers, self.n_patches,patches_center.shape[-1]])

        E_matrix = np.empty(
            (n_patches, n_bins, self._energy_exchange_etc.shape[-1]))
        histogram_out = np.empty((
            n_receivers, n_patches, n_bins,
            self._energy_exchange_etc.shape[-1]))

        receiver_visibility=np.empty((n_receivers,n_patches),dtype=bool)

        for i in range(n_receivers):
            patches_receiver_distance = patches_center - receiver_pos[i]

            receiver_visibility[i] = geometry._check_point2patch_visibility(
                                        eval_point=receiver_pos[i],
                                        patches_center=patches_center,
                                        surf_points=self.walls_points,
                                        surf_normal=self.walls_normal)

            # geometrical weighting
            patch_receiver_energy=form_factor._patch2receiver_energy_universal(
                    receiver_pos[i], patches_points, receiver_visibility[i])

            # access histograms with correct scattering weighting
            receivers_array = np.array(
                [s.cartesian for s in self._brdf_outgoing_directions])

            receiver_idx = get_scattering_data_receiver_index(
                patches_center, receiver_pos[i], receivers_array,
                self._patch_to_wall_ids)

            assert receiver_idx.shape[0] == self.n_patches
            assert len(receiver_idx.shape) == 1

            for k in range(n_patches):
                E_matrix[k,:]= (
                    self._energy_exchange_etc[k,int(receiver_idx[k]),:]
                                                    * patch_receiver_energy[k])

            if propagation_fx:
                # accumulate the patch energies towards the receiver
                # with atmospheric effects (delay, atmospheric attenuation)
                histogram_out[i] = _collect_receiver_energy(
                        E_matrix,
                        np.linalg.norm(patches_receiver_distance, axis=1),
                                    self.speed_of_sound,
                                    self._etc_time_resolution,
                                    air_attenuation=air_attenuation)
            else:
                histogram_out[i] = E_matrix

        return histogram_out

    def set_air_attenuation(self, air_attenuation:pf.FrequencyData):
        """Set air attenuation factor in Np/m.

        Parameters
        ----------
        air_attenuation : pyfar.FrequencyData
            Air attenuation factor in Np/m.

        """
        self._check_set_frequency(air_attenuation.frequencies)
        self._air_attenuation = np.atleast_1d(air_attenuation.freq.squeeze())

    def set_wall_brdf(
            self,
            wall_indexes:list[int],
            brdf:pf.FrequencyData,
            incoming_directions:pf.Coordinates,
            outgoing_directions:pf.Coordinates):
        """Set the wall BRDF representing scattering and absorption.

        For the incoming and outgoing directions, the radius is ignored
        as it only represents a direction. The BRDF assumes an up vector
        of (1, 0, 0) and a normal vector of (0, 0, 1). Therefore, the
        incoming and outgoing directions must not have negative z-components.
        For each wall, the incoming and outgoing directions are rotated
        to align with the given wall's normal vector and up vector.

        Parameters
        ----------
        wall_indexes : list[int]
            List of wall indices for the given BRDF data.
        brdf : pf.FrequencyData
            BRDF data with shape
            (n_incoming_directions, n_outgoing_directions).
        incoming_directions : pf.Coordinates
            Incoming directions of the BRDF data.
        outgoing_directions : pf.Coordinates
            Outgoing directions of the BRDF data.

        """
        assert (incoming_directions.z >= 0).all(), \
            "Sources must be in the positive half space"
        assert (outgoing_directions.z >= 0).all(), \
            "Receivers must be in the positive half space"
        self._check_set_frequency(brdf.frequencies)
        if self._brdf_incoming_directions is None:
            self._brdf_incoming_directions = np.empty(
                (self.n_walls), dtype=pf.Coordinates)
            self._brdf_outgoing_directions = np.empty(
                (self.n_walls), dtype=pf.Coordinates)
            self._brdf_index = np.empty((self.n_walls), dtype=np.int64)
            self._brdf_index.fill(-1)
            self._brdf = []

        for i in wall_indexes:
            incoming_rot, outgoing_rot = _rotate_coords_to_normal(
                self.walls_normal[i], self.walls_up_vector[i],
                incoming_directions, outgoing_directions)
            self._brdf_incoming_directions[i] = incoming_rot
            self._brdf_outgoing_directions[i] = outgoing_rot

        self._brdf.append(brdf.freq*np.pi)
        self._brdf_index[wall_indexes] = len(self._brdf)-1

    def _check_set_frequency(self, frequencies:np.ndarray):
        """Check if the frequency data matches the radiosity object."""
        if self._frequencies is None:
            self._frequencies = frequencies
        else:
            assert self._frequencies.size == frequencies.size, \
                "Number of frequency bins do not match"
            assert (self._frequencies == frequencies).all(), \
                "Frequencies do not match"

    def write(self, filename, compress=True):
        """Write the object to a far file."""
        pf.io.write(filename, compress=compress, **self.to_dict())

    @classmethod
    def from_read(cls, filename):
        """Read the object to a far file."""
        data = pf.io.read(filename)
        for key, value in data.items():
            if isinstance(value, str) and value == 'None':
                data[key] = None
        return cls.from_dict(data)

    def to_dict(self) -> dict:
        """Convert the object to a dictionary."""
        dict_out = {
            'walls_points': self._walls_points,
            'walls_normal': self._walls_normal,
            'walls_up_vector': self._walls_up_vector,
            'patches_points': self._patches_points,
            'n_patches': self._n_patches,
            'patch_to_wall_ids': self._patch_to_wall_ids,
            'visibility_matrix': self._visibility_matrix,
            'visible_patches': self._visible_patches,
            'form_factors': self._form_factors,
            'form_factors_tilde': self._form_factors_tilde,
            'frequencies': self._frequencies,
            'brdf': self._brdf,
            'brdf_index': self._brdf_index,
            'brdf_incoming_directions': self._brdf_incoming_directions,
            'brdf_outgoing_directions': self._brdf_outgoing_directions,
            'patch_2_brdf_outgoing_index': self._patch_2_brdf_outgoing_index,
            'air_attenuation': self._air_attenuation,
            'speed_of_sound': self._speed_of_sound,
            'etc_time_resolution': self._etc_time_resolution,
            'etc_duration': self._etc_duration,
            'distance_patches_to_source': self._distance_patches_to_source,
            'energy_init_source': self._energy_init_source,
            'energy_exchange_etc': self._energy_exchange_etc,
        }
        for key, value in dict_out.items():
            if value is None:
                dict_out[key] = 'None'
            elif isinstance(value, np.ndarray):
                dict_out[key] = value.tolist()
        return dict_out

    def __eq__(self, other):
        """Check for equality of two objects."""
        if not isinstance(other, DirectionalRadiosityFast):
            return False
        return not deepdiff.DeepDiff(self.to_dict(), other.to_dict())


    @classmethod
    def from_dict(cls, input_dict: dict):
        """Create an object from a dictionary. Used for read write."""
        obj = cls(**input_dict)
        return obj

    @property
    def n_bins(self):
        """Return the number of frequency bins."""
        if self._frequencies is None:
            return None
        return self._frequencies.shape[0]

    @property
    def n_walls(self):
        """Return the number of walls."""
        return self._walls_points.shape[0]

    @property
    def n_patches(self):
        """Return the total number of patches."""
        return self._n_patches

    @property
    def form_factors(self):
        """Return the form factor."""
        return self._form_factors

    @property
    def visibility_matrix(self):
        """Return the visibility matrix."""
        return self._visibility_matrix

    @property
    def walls_area(self):
        """Return the area of the walls."""
        return geometry._calculate_area(self._walls_points)

    @property
    def walls_points(self):
        """Return the points of the walls."""
        return self._walls_points

    @property
    def walls_normal(self):
        """Return the normal of the walls."""
        return self._walls_normal

    @property
    def walls_center(self):
        """Return the center of the walls."""
        return geometry._calculate_center(self._walls_points)

    @property
    def walls_up_vector(self):
        """Return the up vector of the walls."""
        return self._walls_up_vector

    @property
    def patches_area(self):
        """Return the area of the patches."""
        return geometry._calculate_area(self._patches_points)

    @property
    def patches_center(self):
        """Return the center of the patches."""
        return geometry._calculate_center(self._patches_points)

    @property
    def patches_size(self):
        """Return the size of the patches."""
        return geometry._calculate_size(self._patches_points)

    @property
    def patches_points(self):
        """Return the points of the patches."""
        return self._patches_points

    @property
    def patches_normal(self):
        """Return the normal of the patches."""
        return self._walls_normal[self._patch_to_wall_ids]

    @property
    def speed_of_sound(self):
        """Return the speed of sound in m/s."""
        return self._speed_of_sound


def _rotate_coords_to_normal(
        wall_normal:np.ndarray, wall_up_vector:np.ndarray,
        sources:pf.Coordinates, receivers:pf.Coordinates):
    """Rotate the coordinates to the normal vector."""
    o1 = pf.Orientations.from_view_up(
        wall_normal, wall_up_vector)
    o2 = pf.Orientations.from_view_up([0, 0, 1], [1, 0, 0])
    o_diff = o1.inv()*o2
    euler = o_diff.as_euler('xyz', True).flatten()
    receivers_cp = receivers.copy()
    receivers_cp.rotate('xyz', euler)
    receivers_cp.radius = 1
    sources_cp = sources.copy()
    sources_cp.rotate('xyz', euler)
    sources_cp.radius = 1
    return sources_cp, receivers_cp

def _add_directional(
        energy_0, source_position: np.ndarray,
        patches_center: np.ndarray, n_bins:float, patch_to_wall_ids:np.ndarray,
        sources: np.ndarray, receivers: np.ndarray,
        scattering: np.ndarray, scattering_index: np.ndarray):
    """Add scattering and absorption to the initial energy from the source.

    Parameters
    ----------
    energy_0 : np.ndarray
        energy of all patches of shape (n_patches, n_bins)
    source_position : np.ndarray
        source position of shape (3,)
    patches_center : np.ndarray
        center of all patches of shape (n_patches, 3)
    n_bins : float
        number of frequency bins.
    patch_to_wall_ids : np.ndarray
        indexes from each patch to the wall of shape (n_patches)
    sources : np.ndarray
        source positions of shape (n_walls, n_sources, 3)
    receivers : np.ndarray
        receiver positions of shape (n_walls, n_receivers, 3)
    scattering : np.ndarray
        scattering data of shape (n_walls, n_sources, n_receivers, n_bins)
    scattering_index : np.ndarray
        mapping from the wall id to scattering database index (n_walls)

    Returns
    -------
    energy : np.ndarray
        energy of all patches of shape (n_patches, n_directions, n_bins)

    """
    n_patches = patches_center.shape[0]
    n_directions = receivers.shape[1]
    energy_0_directivity = np.zeros((n_patches, n_directions, n_bins))
    for i in prange(n_patches):
        wall_id_i = int(patch_to_wall_ids[i])
        scattering_factor = get_scattering_data_source(
            source_position, patches_center[i],
            sources, wall_id_i, scattering, scattering_index)

        energy_0_directivity[i, :, :] = energy_0[i] \
            * np.real(scattering_factor)

    return energy_0_directivity


def _energy_exchange_init_energy(
        n_samples, energy_0_directivity, distance_0,
        speed_of_sound, histogram_time_resolution):
    """Calculate energy exchange between patches.

    Parameters
    ----------
    n_samples : int
        number of samples of the histogram.
    energy_0_directivity : np.ndarray
        init energy of all patches of shape (n_patches, n_directions, n_bins)
    distance_0 : np.ndarray
        distance from the source to all patches of shape (n_patches)
    speed_of_sound : float
        speed of sound in m/s.
    histogram_time_resolution : float
        time resolution of the histogram in s.

    Returns
    -------
    E_matrix_total : np.ndarray
        energy of all patches of shape
        (n_patches, n_directions, n_bins, n_samples)

    """
    n_patches = energy_0_directivity.shape[0]
    n_directions = energy_0_directivity.shape[1]
    n_bins = energy_0_directivity.shape[2]
    E_matrix_total = np.zeros((n_patches, n_directions, n_bins, n_samples))
    for i in prange(n_patches):
        n_delay_samples = int(
            distance_0[i]/speed_of_sound/histogram_time_resolution)
        E_matrix_total[i, :, :, n_delay_samples] += energy_0_directivity[i]
    return E_matrix_total


def _energy_exchange(
        n_samples, energy_0_directivity, distance_0, distance_ij,
        form_factors_tilde, patch_2_out_directions,
        speed_of_sound, histogram_time_resolution, max_order, visible_patches):
    """Calculate energy exchange between patches.

    Parameters
    ----------
    n_samples : int
        number of samples of the histogram.
    energy_0_directivity : np.ndarray
        init energy of all patches of shape (n_patches, n_directions, n_bins)
    distance_0 : np.ndarray
        distance from the source to all patches of shape (n_patches)
    distance_ij : np.ndarray
        distance between all patches of shape (n_patches, n_patches)
    form_factors_tilde : np.ndarray
        form factors of shape (n_patches, n_patches, n_directions, n_bins)
    patch_2_out_directions: np.ndarray
        patchwise map of patch centers to
        scattering outgoing directions of shape (n_patches,n_patches)
    speed_of_sound : float
        speed of sound in m/s.
    histogram_time_resolution : float
        time resolution of the histogram in s.
    max_order : int
        maximum order of reflections.
    visible_patches : np.ndarray
        indexes of all visible patches of shape (n_visible, 2)

    Returns
    -------
    E_matrix_total : np.ndarray
        energy of all patches of shape
        (n_patches, n_directions, n_bins, n_samples)

    """
    n_patches = form_factors_tilde.shape[0]
    n_directions = form_factors_tilde.shape[2]
    n_bins = energy_0_directivity.shape[2]
    form_factors_tilde = form_factors_tilde[..., np.newaxis]
    E_matrix_total  = _energy_exchange_init_energy(
        n_samples, energy_0_directivity, distance_0, speed_of_sound,
        histogram_time_resolution)
    E_matrix = np.zeros((2, n_patches, n_directions, n_bins, n_samples))
    E_matrix[0] += E_matrix_total
    if max_order == 0:
        return E_matrix_total
    for k in range(max_order):
        current_index = (1+k) % 2
        E_matrix[current_index, :, :, :] = 0
        for ii in range(visible_patches.shape[0]):
            for jj in range(2):
                if jj == 0:
                    i = visible_patches[ii, 0]
                    j = visible_patches[ii, 1]
                else:
                    j = visible_patches[ii, 0]
                    i = visible_patches[ii, 1]

                dir_id = patch_2_out_directions[i,j]

                n_delay_samples = int(
                    distance_ij[i, j]/speed_of_sound/histogram_time_resolution)
                if n_delay_samples > 0:
                    E_matrix[current_index, j, :, :, n_delay_samples:] += \
                        form_factors_tilde[i, j] * E_matrix[
                            current_index-1, i, dir_id, :, :-n_delay_samples]
                else:
                    E_matrix[current_index, j, :, :, :] += form_factors_tilde[
                        i, j] * E_matrix[current_index-1, i, dir_id, :, :]
        E_matrix_total += E_matrix[current_index]
    return E_matrix_total


def _collect_receiver_energy(
        E_matrix_total, patch_receiver_distance,
        speed_of_sound, histogram_time_resolution, air_attenuation):
    """Collect the energy at the receiver.

    Parameters
    ----------
    E_matrix_total : np.ndarray
        energy of all patches of shape
        (n_patches, n_directions, n_bins, n_samples)
    patch_receiver_distance : np.ndarray
        distance from the patch to the receiver of shape (n_patches)
    speed_of_sound : float
        speed of sound in m/s.
    histogram_time_resolution : float
        time resolution of the histogram in s.
    air_attenuation : np.ndarray
        air attenuation factor for each frequency bin of shape (n_bins)

    Returns
    -------
    ir : np.ndarray
        impulse response of shape (n_samples, n_bins)

    """
    E_mat_out = np.zeros_like(E_matrix_total)
    n_patches = E_matrix_total.shape[0]
    n_bins = E_matrix_total.shape[1]

    for i in prange(n_patches):
        n_delay_samples = int(np.ceil(
            patch_receiver_distance[i]/speed_of_sound/histogram_time_resolution))
        for j in range(n_bins):
            E_mat_out[i,j] = np.roll(
                E_matrix_total[i,j]*np.exp(-air_attenuation[j]*patch_receiver_distance[i]),
                n_delay_samples)

    return E_mat_out

def _form_factors_with_directivity(
        visibility_matrix, form_factors, n_bins, patches_center,
        patches_area, air_attenuation,
        absorption, absorption_index, patch_to_wall_ids,
        scattering, scattering_index, sources, receivers):
    """Calculate the form factors with directivity."""
    n_patches = patches_center.shape[0]
    form_factors_tilde = np.zeros((n_patches, n_patches, n_patches, n_bins))
    # loop over previous patches, current and next patch

    for ii in prange(n_patches**3):
        h = ii % n_patches
        i = int(ii/n_patches) % n_patches
        j = int(ii/n_patches**2) % n_patches
        visible_hi = visibility_matrix[
            h, i] if h < i else visibility_matrix[i, h]
        visible_ij = visibility_matrix[
            i, j] if i < j else visibility_matrix[j, i]
        if visible_hi and visible_ij:
            difference_receiver = patches_center[i]-patches_center[j]
            wall_id_i = int(patch_to_wall_ids[i])
            difference_receiver /= np.linalg.norm(difference_receiver)
            ff = form_factors[i, j] if i<j else (form_factors[j, i]
                                                 *patches_area[i]/patches_area[j])

            distance = np.linalg.norm(difference_receiver)
            form_factors_tilde[h, i, j, :] = ff
            if air_attenuation is not None:
                form_factors_tilde[h, i, j, :] = form_factors_tilde[
                    h, i, j, :] * np.exp(-air_attenuation * distance)

            if scattering is not None:
                scattering_factor = get_scattering_data(
                    patches_center[h], patches_center[i], patches_center[j],
                    sources, receivers, wall_id_i,
                    scattering, scattering_index)
                form_factors_tilde[h, i, j, :] = form_factors_tilde[
                    h, i, j, :] * scattering_factor

            if absorption is not None:
                source_wall_idx = absorption_index[wall_id_i]
                form_factors_tilde[h, i, j, :] = form_factors_tilde[
                    h, i, j, :] * (1-absorption[source_wall_idx])

    return form_factors_tilde


def _form_factors_with_directivity_dim(
        visibility_matrix, form_factors, n_bins, patches_center,
        patches_area,
        air_attenuation,
        patch_to_wall_ids,
        scattering, scattering_index, sources, receivers):
    """Calculate the form factors with directivity."""
    n_patches = patches_center.shape[0]
    n_directions = receivers.shape[1] if receivers is not None else 1
    form_factors_tilde = np.zeros((n_patches, n_patches, n_directions, n_bins))
    # loop over previous patches, current and next patch

    for ii in prange(n_patches**2):
        i = ii % n_patches
        j = int(ii/n_patches) % n_patches
        visible_ij = visibility_matrix[
            i, j] if i < j else visibility_matrix[j, i]
        if visible_ij:
            difference_receiver = patches_center[i]-patches_center[j]
            wall_id_i = int(patch_to_wall_ids[i])
            difference_receiver /= np.linalg.norm(difference_receiver)
            ff = form_factors[i, j] if i<j else (form_factors[j, i]
                                                 *patches_area[j]/patches_area[i])

            distance = np.linalg.norm(difference_receiver)
            form_factors_tilde[i, j, :, :] = ff
            if air_attenuation is not None:
                form_factors_tilde[i, j, :, :] = form_factors_tilde[
                    i, j, :, :] * np.exp(-air_attenuation * distance)

            if scattering is not None:
                scattering_factor = get_scattering_data_source(
                    patches_center[i], patches_center[j],
                    sources, wall_id_i,
                    scattering, scattering_index)
                form_factors_tilde[i, j, :, :] = form_factors_tilde[
                    i, j, :, :] * scattering_factor

    return form_factors_tilde


## scattering data access

def get_scattering_data_receiver_index(
        pos_i:np.ndarray, pos_j:np.ndarray,
        receivers:np.ndarray, wall_id_i:np.ndarray,
        ):
    """Get scattering receiver index based on current and next position.

    Parameters
    ----------
    pos_i : np.ndarray
        current position of shape (n,3)
    pos_j : np.ndarray
        next position of shape (3)
    receivers : np.ndarray
        receiver directions of all walls of shape (n_walls, n_receivers, 3)
    wall_id_i : np.ndarray
        current wall id to get write directional data

    Returns
    -------
    scattering_factor: float
        scattering factor from directivity

    """
    n_patches = pos_i.shape[0] if pos_i.ndim > 1 else 1
    receiver_idx = np.empty((n_patches), dtype=np.int64)

    for i in range(n_patches):
        difference_receiver = pos_j-pos_i[i]

        difference_receiver /= np.linalg.norm(
            difference_receiver)
        receiver_idx[i] = np.argmin(np.sum(
            (receivers[wall_id_i[i], :]-difference_receiver)**2, axis=-1),
            axis=-1)

    return receiver_idx


def get_scattering_data(
        pos_h:np.ndarray, pos_i:np.ndarray, pos_j:np.ndarray,
        sources:np.ndarray, receivers:np.ndarray, wall_id_i:np.ndarray,
        scattering:np.ndarray, scattering_index:np.ndarray):
    """Get scattering data depending on previous, current and next position.

    Parameters
    ----------
    pos_h : np.ndarray
        previous position of shape (3)
    pos_i : np.ndarray
        current position of shape (3)
    pos_j : np.ndarray
        next position of shape (3)
    sources : np.ndarray
        source directions of all walls of shape (n_walls, n_sources, 3)
    receivers : np.ndarray
        receiver directions of all walls of shape (n_walls, n_receivers, 3)
    wall_id_i : np.ndarray
        current wall id to get write directional data
    scattering : np.ndarray
        scattering data of shape (n_scattering, n_sources, n_receivers, n_bins)
    scattering_index : np.ndarray
        index of the scattering data of shape (n_walls)

    Returns
    -------
    scattering_factor: float
        scattering factor from directivity

    """
    difference_source = pos_h-pos_i
    difference_receiver = pos_i-pos_j

    difference_source /= np.linalg.norm(difference_source)
    difference_receiver /= np.linalg.norm(difference_receiver)
    source_idx = np.argmin(np.sum(
        (sources[wall_id_i, :, :]-difference_source)**2, axis=-1))
    receiver_idx = np.argmin(np.sum(
        (receivers[wall_id_i, :]-difference_receiver)**2, axis=-1))
    return scattering[scattering_index[wall_id_i],
        source_idx, receiver_idx, :]


def get_scattering_data_source(
        pos_h:np.ndarray, pos_i:np.ndarray,
        sources:np.ndarray, wall_id_i:np.ndarray,
        scattering:np.ndarray, scattering_index:np.ndarray):
    """Get scattering data depending on previous, current position.

    Parameters
    ----------
    pos_h : np.ndarray
        previous position of shape (3)
    pos_i : np.ndarray
        current position of shape (3)
    sources : np.ndarray
        source directions of all walls of shape (n_walls, n_sources, 3)
    wall_id_i : np.ndarray
        current wall id to get write directional data
    scattering : np.ndarray
        scattering data of shape (n_scattering, n_sources, n_receivers, n_bins)
    scattering_index : np.ndarray
        index of the scattering data of shape (n_walls)

    Returns
    -------
    scattering_factor: float
        scattering factor from directivity

    """
    difference_source = pos_h-pos_i
    difference_source /= np.linalg.norm(difference_source)
    source_idx = np.argmin(np.sum(
        (sources[wall_id_i, :, :]-difference_source)**2, axis=-1))
    return scattering[scattering_index[wall_id_i], source_idx]

if numba is not None:
    _add_directional = numba.njit(parallel=True)(_add_directional)
    _energy_exchange_init_energy = numba.njit()(_energy_exchange_init_energy)
    _collect_receiver_energy = numba.njit()(_collect_receiver_energy)
    _energy_exchange = numba.njit()(_energy_exchange)
    _form_factors_with_directivity_dim = numba.njit(parallel=True)(
        _form_factors_with_directivity_dim)
    _form_factors_with_directivity = numba.njit(parallel=True)(
        _form_factors_with_directivity)
    get_scattering_data_receiver_index = numba.njit()(
        get_scattering_data_receiver_index)
    get_scattering_data = numba.njit()(get_scattering_data)
    get_scattering_data_source = numba.njit()(get_scattering_data_source)






"""Test radiosity module."""
import numpy as np
import numpy.testing as npt
import sparrowpy.geometry as geo
import sparrowpy as sp
from sparrowpy.sound_object import SoundSource


def test_patches():
    """Test Patches class."""
    poly = geo.Polygon(
        [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]], [0, 1, 0], [0, 0, 1])
    patches = sp.PatchesKang(poly, 0.1, [], 0)
    assert len(patches.patches) == 100
    for patch in patches.patches:
        # check if area is always same
        dimensions = np.max(patch.pts, axis=0) - np.min(patch.pts, axis=0)
        dimensions = dimensions[dimensions > 0]
        area = dimensions[0]*dimensions[1]
        npt.assert_almost_equal(area, 0.01)
    npt.assert_almost_equal(patches.patches[0].pts, poly.pts/10)


def test_patches_dim2():
    """Test Patches class with 2D polygon."""
    poly = geo.Polygon(
        [[0, 0, 0], [0, 1, 0], [0, 1, 1], [0, 0, 1]], [0, 1, 0], [1, 0, 0])
    patches = sp.PatchesKang(poly, 0.1, [], 0)
    assert len(patches.patches) == 100
    for patch in patches.patches:
        # check if area is always same
        dimensions = np.max(patch.pts, axis=0) - np.min(patch.pts, axis=0)
        dimensions = dimensions[dimensions > 0]
        area = dimensions[0]*dimensions[1]
        npt.assert_almost_equal(area, 0.01)
    npt.assert_almost_equal(patches.patches[0].pts, poly.pts/10)


def test_form_factor():
    """Test form_factor function."""
    poly1 = geo.Polygon(
        [[0, 0, 0], [0, 1, 0], [0, 1, 1], [0, 0, 1]], [0, 1, 0], [1, 0, 0])
    patches1 = sp.PatchesKang(poly1, 0.2, [1], 0)
    poly2 = geo.Polygon(
        [[2, 0, 0], [2, 1, 0], [2, 1, 1], [2, 0, 1]], [2, 1, 0], [1, 0, 0])
    patches2 = sp.PatchesKang(poly2, 0.2, [0], 1)
    patches1.calculate_form_factor([patches1, patches2])


def test_calculate_form_factors():
    """Test calculate_form_factors function."""
    # set geometry
    X = 8
    Y = 4
    Z = 4
    patch_size = 2

    ground = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, Y, 0], [0, Y, 0]], [1, 0, 0], [0, 0, 1])
    A_wall = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, 0, Z], [0, 0, Z]], [1, 0, 0], [0, 1, 0])
    B_wall = geo.Polygon(
        [[0, Y, 0], [X, Y, 0], [X, Y, Z], [0, Y, Z]], [1, 0, 0], [0, -1, 0])

    ground_patches = sp.PatchesKang(ground, patch_size, [1, 2], 0)
    A_wall_patches = sp.PatchesKang(A_wall, patch_size, [0, 2], 1)
    B_wall_patches = sp.PatchesKang(B_wall, patch_size, [0, 1], 2)

    patches_list = [ground_patches, A_wall_patches, B_wall_patches]

    # calculate form factor
    A_wall_patches.calculate_form_factor(patches_list)


def test_init_energy_matrix():
    """Test init_energy_matrix function."""
    X = 8
    Y = 4
    patch_size = 2
    ground = geo.Polygon(
        [[0, 0, 0], [X, 0, 0], [X, Y, 0], [0, Y, 0]], [1, 0, 0], [0, 0, 1])
    ground_patches = sp.PatchesKang(ground, patch_size, [1, 2], 0)
    max_order_k = 2
    ir_length_s = 1
    sampling_rate = 1000
    source = SoundSource([2, 2, 1], [0, 1, 0], [0, 0, 1])
    ground_patches.init_energy_exchange(
        max_order_k, ir_length_s, source, sampling_rate=sampling_rate,
        speed_of_sound=343)
    assert ground_patches.E_n_samples == 1000
    assert ground_patches.E_sampling_rate == sampling_rate
    npt.assert_array_equal(
        ground_patches.E_matrix.shape, (1, max_order_k+1, 8, 1000))
